"""Unit tests for the one.webclient module"""
import unittest
from unittest import mock
import random
import os
import io
import one.webclient as wc
import one.params
import tempfile
import shutil
import requests
import json
import logging
from datetime import datetime, timedelta
from uuid import UUID

from iblutil.io import hashfile
import iblutil.io.params as iopar

from . import OFFLINE_ONLY, TEST_DB_1, TEST_DB_2
from . import util

par = one.params.get(silent=True)

# Init connection to the database
ac = wc.AlyxClient(**TEST_DB_1)
# Remove /public from data server url
if 'public' in ac._par.HTTP_DATA_SERVER:
    ac._par = ac._par.set('HTTP_DATA_SERVER', ac._par.HTTP_DATA_SERVER.rsplit('/', 1)[0])


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestAuthentication(unittest.TestCase):
    """Tests for AlyxClient authentication, token storage, login/out methods and user prompts"""
    def setUp(self) -> None:
        self.ac = wc.AlyxClient(**TEST_DB_2)

    def test_authentication(self):
        """Test for AlyxClient.authenticate and AlyxClient.is_logged_in property"""
        ac = self.ac
        self.assertTrue(ac.is_logged_in)
        ac.logout()
        self.assertFalse(ac.is_logged_in)
        # Check token removed from cache
        cached_token = getattr(one.params.get(TEST_DB_2['base_url']), 'TOKEN', {})
        self.assertFalse(TEST_DB_2['username'] in cached_token)
        # Test with pars set
        login_keys = {'ALYX_LOGIN', 'ALYX_PWD'}
        if not set(ac._par.as_dict().keys()) >= login_keys:
            for k, v in zip(sorted(login_keys), (TEST_DB_2['username'], TEST_DB_2['password'])):
                ac._par = ac._par.set(k, v)
        with mock.patch('builtins.input') as mock_input:
            ac.authenticate()
            mock_input.assert_not_called()
        self.assertTrue(ac.is_logged_in)
        # Test using input args
        ac._par = iopar.from_dict({k: v for k, v in ac._par.as_dict().items()
                                   if k not in login_keys})
        ac.logout()
        with mock.patch('builtins.input') as mock_input:
            ac.authenticate(TEST_DB_2['username'], TEST_DB_2['password'], cache_token=False)
            mock_input.assert_not_called()
        # Check token not saved in cache
        cached_token = getattr(one.params.get(TEST_DB_2['base_url']), 'TOKEN', {})
        self.assertFalse(TEST_DB_2['username'] in cached_token)
        # Test user prompts
        ac.logout()
        ac.silent = False
        with mock.patch('builtins.input', return_value=TEST_DB_2['username']),\
             mock.patch('one.webclient.getpass', return_value=TEST_DB_2['password']):
            ac.authenticate(cache_token=True)
        self.assertTrue(ac.is_logged_in)
        # Check token saved in cache
        ac.authenticate(cache_token=True)
        cached_token = getattr(one.params.get(TEST_DB_2['base_url']), 'TOKEN', {})
        self.assertTrue(TEST_DB_2['username'] in cached_token)
        # Check force flag
        with mock.patch('one.webclient.getpass', return_value=TEST_DB_2['password']) as mock_pwd:
            ac.authenticate(cache_token=True, force=True)
            mock_pwd.assert_called()
        # Check non-silent double logout
        ac.logout()
        ac.logout()  # Shouldn't complain

    def test_auth_methods(self):
        """Test behaviour when calling AlyxClient._generic_request when logged out"""
        # Check that authentication happens when making a logged out request
        self.ac.logout()
        assert not self.ac.is_logged_in
        # Set pars for auto login
        login_keys = {'ALYX_LOGIN', 'ALYX_PWD'}
        if not set(self.ac._par.as_dict().keys()) >= login_keys:
            for k, v in zip(sorted(login_keys), (TEST_DB_2['username'], TEST_DB_2['password'])):
                self.ac._par = self.ac._par.set(k, v)

        # Test generic request
        self.ac._generic_request(requests.get, '/sessions?user=Hamish', clobber=True)
        self.assertTrue(self.ac.is_logged_in)

        # Test behaviour when token invalid
        self.ac._token['token'] = '1NVAL1DT0K3N'
        self.ac._headers['Authorization'] = 'Token ' + self.ac._token['token']
        self.ac._generic_request(requests.get, '/sessions?user=Hamish', clobber=True)
        self.assertTrue(self.ac.is_logged_in)

        # Test download cache tables
        self.ac.logout()
        assert not self.ac.is_logged_in
        url = self.ac.get('cache/info')['location']
        self.ac.download_cache_tables(url)
        self.assertTrue(self.ac.is_logged_in)

    def test_auth_errors(self):
        """Test behaviour when authentication fails"""
        self.ac.logout()  # Make sure logged out
        with self.assertRaises(requests.HTTPError) as ex:
            self.ac.authenticate(password='wrong_pass')
            self.assertTrue('user = intbrainlab' in str(ex))
            self.assertFalse('wrong_pass' in str(ex))
        # Check behaviour when connection error occurs (should mention firewall settings)
        with mock.patch('one.webclient.requests.post', side_effect=requests.ConnectionError),\
             self.assertRaises(ConnectionError) as ex:
            self.ac.authenticate()
            self.assertTrue('firewall' in str(ex))
        # Check behaviour when server error occurs
        rep = requests.Response()
        rep.status_code = 500
        with mock.patch('one.webclient.requests.post', return_value=rep),\
             self.assertRaises(requests.HTTPError):
            self.ac.authenticate()


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestJsonFieldMethods(unittest.TestCase):
    """Tests for AlyxClient methods that modify the JSON field of a REST endpoint.

    These tests are over-engineered in order to test Alyx Django queries with JSON fields.
    Django queries are also tested in TestRemote.test_search.
    """
    def setUp(self):
        self.ac = wc.AlyxClient(**TEST_DB_1, cache_rest=None)

        # Create new subject and two new sessions
        name = '0A' + str(random.randint(0, 10000))
        self.subj = self.ac.rest('subjects', 'create', data={'nickname': name, 'lab': 'cortexlab'})
        sessions = [self.ac.rest('sessions', 'create', data={
            'subject': name,
            'start_time': datetime.isoformat(datetime.now()),
            'number': random.randint(1, 999),
            'type': 'Experiment',
            'users': [TEST_DB_1['username']],
        }) for _ in range(2)]

        self.eids = [x['url'].split('/')[-1] for x in sessions]
        self.endpoint = 'sessions'
        self.field_name = 'extended_qc'
        # We filter by key value so we use randint to avoid race condition in concurrent tests
        i = str(random.randint(0, 10000))
        self.data_dict = {'low_' + i: 0, 'high_' + i: 1}

    def _json_field_write(self):
        written1 = self.ac.json_field_write(
            self.endpoint, self.eids[0], self.field_name, self.data_dict
        )
        written2 = self.ac.json_field_write(
            self.endpoint, self.eids[1], self.field_name, self.data_dict
        )
        self.assertTrue(written1 == written2)
        self.assertTrue(written1 == self.data_dict)
        data_field = next(filter(lambda x: x.startswith('low'), self.data_dict))
        url = f'/{self.endpoint}?&{self.field_name}={data_field}__lt,0.5'
        sess_dict = self.ac.get(url, expires=True)
        self.assertTrue(len(sess_dict) == 2)

    def _json_field_update(self):
        data_field = next(filter(lambda x: x.startswith('low'), self.data_dict))
        modified = self.ac.json_field_update(
            self.endpoint, self.eids[0], self.field_name, {data_field: 0.6}
        )
        self.assertCountEqual(modified.keys(), self.data_dict.keys())
        url = f'/{self.endpoint}?&{self.field_name}={data_field}__lt,0.5'
        self.assertTrue(len(self.ac.get(url, expires=True)) == 1)

    def _json_field_remove_key(self):
        eid = self.eids[1]
        data_field = next(filter(lambda x: x.startswith('hi'), self.data_dict))
        url = f'/{self.endpoint}?&{self.field_name}={data_field}__gte,0.5'
        pre_delete = self.ac.get(url, expires=True)
        self.assertTrue(len(pre_delete) == 2)
        deleted = self.ac.json_field_remove_key(self.endpoint, eid, self.field_name, data_field)
        self.assertTrue(data_field not in deleted)
        post_delete = self.ac.get(url, expires=True)
        self.assertTrue(len(post_delete) == 1)

    def _json_field_delete(self):
        data_field = next(filter(lambda x: x.startswith('hi'), self.data_dict))
        deleted = self.ac.json_field_delete(self.endpoint, self.eids[1], self.field_name)
        self.assertTrue(deleted is None)
        url = f'/{self.endpoint}?&{self.field_name}={data_field}__gte,0.5'
        ses = self.ac.get(url, expires=True)
        self.assertTrue(len(ses) == 1)

    def test_json_methods(self):
        """Test for AlyxClient.json_field* methods (write, update, remove_key and delete)"""
        self._json_field_write()
        self._json_field_update()
        self._json_field_remove_key()
        self._json_field_delete()

    def test_empty(self):
        """Test for AlyxClient.json_field* methods when JSON field is empty"""
        eid = self.eids[0]
        # Check behaviour when fields are empty
        self.ac.rest(self.endpoint, 'partial_update', id=eid, data={self.field_name: None})
        # Should return None as no keys exist
        modified = self.ac.json_field_remove_key(self.endpoint, eid, self.field_name, 'foo')
        self.assertIsNone(modified)
        # Should return data
        data = {'some': 0.6}
        modified = self.ac.json_field_update(self.endpoint, eid, self.field_name, data)
        self.assertTrue(modified == data)
        # Should warn if key not in dict
        with self.assertLogs(logging.getLogger('one.webclient'), logging.WARNING):
            self.ac.json_field_remove_key(self.endpoint, eid, self.field_name, 'foo')
        # Check behaviour when fields not a dict
        data = {self.field_name: json.dumps(data)}
        self.ac.rest(self.endpoint, 'partial_update', id=eid, data=data)
        # Update field
        with self.assertLogs(logging.getLogger('one.webclient'), logging.WARNING):
            modified = self.ac.json_field_update(self.endpoint, eid, self.field_name, data)
        self.assertEqual(data[self.field_name], modified)
        # Remove key
        with self.assertLogs(logging.getLogger('one.webclient'), logging.WARNING):
            modified = self.ac.json_field_remove_key(self.endpoint, eid, self.field_name)
        self.assertIsNone(modified)

    def tearDown(self):
        self.ac.rest('subjects', 'delete', id=self.subj['nickname'])


class TestRestCache(unittest.TestCase):
    """Tests for REST caching system, the cache decorator and cache flags"""
    def setUp(self):
        util.setup_test_params()  # Ensure test alyx set up
        util.setup_rest_cache(ac.cache_dir)  # Copy rest cache fixtures
        self.query = '/insertions/b529f2d8-cdae-4d59-aba2-cbd1b5572e36'
        self.tempdir = util.set_up_env()
        self.addCleanup(self.tempdir.cleanup)
        one.webclient.datetime = _FakeDateTime
        _FakeDateTime._now = None
        self.cache_dir = ac.cache_dir.joinpath('.rest')
        self.default_expiry = ac.default_expiry
        self.cache_mode = ac.cache_mode

    def test_loads_cached(self):
        """Test for one.webclient._cache_response decorator, checks returns cached result"""
        # Check returns cache
        wrapped = wc._cache_response(lambda *args: self.assertTrue(False))
        client = ac  # Bunch({'base_url': 'https://test.alyx.internationalbrainlab.org'})
        res = wrapped(client, requests.get, self.query)
        self.assertEqual(res['id'], self.query.split('/')[-1])

    def test_expired_cache(self):
        """Test behaviour when cached REST query is expired"""
        # Checks expired
        wrapped = wc._cache_response(lambda *args: 'called')
        _FakeDateTime._now = datetime.fromisoformat('3001-01-01')
        res = wrapped(ac, requests.get, self.query)
        self.assertTrue(res == 'called')

    def test_caches_response(self):
        """Test caches query response before returning"""
        # Default expiry time
        ac.default_expiry = timedelta(minutes=1)
        wrapped = wc._cache_response(lambda *args: 'called')
        _FakeDateTime._now = datetime(2021, 5, 13)  # Freeze time
        res = wrapped(ac, requests.get, '/endpoint?id=5')
        self.assertTrue(res == 'called')

        # Check cache file created
        filename = '64b5b3476c015e04ee7c4753606b5e967325d34a'
        cache_file = self.cache_dir / filename
        self.assertTrue(cache_file.exists())
        with open(cache_file, 'r') as f:
            q, when = json.load(f)
        self.assertEqual('called', q)
        self.assertEqual(when, '2021-05-13T00:01:00')

    def test_cache_mode(self):
        """Test for AlyxClient.cache_mode property"""
        # With cache mode off, wrapped method should be called even in presence of valid cache
        ac.cache_mode = None  # cache nothing
        wrapped = wc._cache_response(lambda *args: 'called')
        res = wrapped(ac, requests.get, self.query)
        self.assertTrue(res == 'called')

    def test_expiry_param(self):
        """Test for expires kwarg in one.webclient._cache_response decorator"""
        # Check expiry param
        wrapped = wc._cache_response(lambda *args: '123')
        res = wrapped(ac, requests.get, '/endpoint?id=5', expires=True)
        self.assertTrue(res == '123')

        # A second call should yield a new response as cache immediately expired
        wrapped = wc._cache_response(lambda *args: '456')
        res = wrapped(ac, requests.get, '/endpoint?id=5', expires=False)
        self.assertTrue(res == '456')

        # With clobber=True the cache should be overwritten
        wrapped = wc._cache_response(lambda *args: '789')
        res = wrapped(ac, requests.get, '/endpoint?id=5', clobber=True)
        self.assertTrue(res == '789')

    def test_cache_returned_on_error(self):
        """Test behaviour when connection error occurs and cached response exists"""
        func = mock.Mock(side_effect=requests.ConnectionError())
        wrapped = wc._cache_response(func)
        _FakeDateTime._now = datetime.fromisoformat('3001-01-01')  # Expired
        with self.assertWarns(RuntimeWarning):
            res = wrapped(ac, requests.get, self.query)
        self.assertEqual(res['id'], self.query.split('/')[-1])

        # With clobber=True exception should be raised
        with self.assertRaises(requests.ConnectionError):
            wrapped(ac, requests.get, self.query, clobber=True)

    def test_clear_cache(self):
        """Test for AlyxClient.clear_rest_cache"""
        assert any(self.cache_dir.glob('*'))
        ac.clear_rest_cache()
        self.assertFalse(any(self.cache_dir.glob('*')))

    def tearDown(self) -> None:
        ac.cache_mode = self.cache_mode
        ac.default_expiry = self.default_expiry


class _FakeDateTime(datetime):
    _now = None

    @staticmethod
    def now(*args, **kwargs):
        return _FakeDateTime._now or datetime.now(*args, **kwargs)


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestDownloadHTTP(unittest.TestCase):

    def setUp(self):
        self.ac = ac
        self.test_data_uuid = '40af4a49-1b9d-45ec-b443-a151c010ea3c'  # OpenAlyx dataset

    def test_paginated_request(self):
        """Check that paginated response object is returned upon making large queries"""
        rep = self.ac.rest('datasets', 'list')
        self.assertTrue(isinstance(rep, one.webclient._PaginatedResponse))
        self.assertTrue(len(rep) > 250)
        # This fails when new records are added/removed from the remote db while iterating
        # self.assertTrue(len([_ for _ in rep]) == len(rep))

        # Test what happens when list changes between paginated requests
        name = '0A' + str(random.randint(0, 10000))
        # Add subject between calls
        rep = self.ac.rest('subjects', 'list', limit=5, no_cache=True)
        s = self.ac.rest('subjects', 'create', data={'nickname': name, 'lab': 'cortexlab'})
        self.addCleanup(self.ac.rest, 'subjects', 'delete', id=s['nickname'])
        with self.assertWarns(RuntimeWarning):
            _ = rep[10]

    def test_update_url_params(self):
        url = f'{self.ac.base_url}/sessions?param1=foo&param2=&limit=5&param3=bar'
        expected = f'{self.ac.base_url}/sessions?param1=foo&limit=10&param3=bar&param4=baz'
        self.assertEqual(expected, wc.update_url_params(url, {'limit': 10, 'param4': 'baz'}))
        # Without pars
        url = url.split('?')[0]
        self.assertEqual(url, wc.update_url_params(url, {}))
        # With lists
        expected = f'{url}?foo=bar&foo=baz'
        self.assertEqual(expected, wc.update_url_params(url, {'foo': ['bar', 'baz']}))

    def test_generic_request(self):
        a = self.ac.get('/labs')
        b = self.ac.get('labs')
        self.assertEqual(a, b)

    def test_rest_endpoint_write(self):
        # test object creation and deletion with weighings
        wa = {'subject': 'flowers',
              'date_time': '2018-06-30T12:34:57',
              'weight': 22.2,
              'user': 'olivier'
              }
        a = self.ac.rest('weighings', 'create', data=wa)
        b = self.ac.rest('weighings', 'read', id=a['url'])
        self.assertEqual(a, b)
        self.ac.rest('weighings', 'delete', id=a['url'])
        # test patch object with subjects
        data = {'birth_date': '2018-04-01',
                'death_date': '2018-09-10'}
        sub = self.ac.rest('subjects', 'partial_update', id='flowers', data=data)
        self.assertEqual(sub['birth_date'], data['birth_date'])
        self.assertEqual(sub['death_date'], data['death_date'])
        data = {'birth_date': '2018-04-02',
                'death_date': '2018-09-09'}
        sub = self.ac.rest('subjects', 'partial_update', id='flowers', data=data)
        self.assertEqual(sub['birth_date'], data['birth_date'])
        self.assertEqual(sub['death_date'], data['death_date'])

    def test_rest_endpoint_read_only(self):
        """Test AlyxClient.rest method with 'list' and 'read' actions"""
        # tests that non-existing endpoints /actions are caught properly
        with self.assertRaises(ValueError):
            self.ac.rest(url='turlu', action='create')
        with self.assertRaises(ValueError):
            self.ac.rest(url='sessions', action='turlu')
        # test with labs : get
        a = self.ac.rest('labs', 'list')
        self.assertTrue(len(a) >= 3)
        b = self.ac.rest('/labs', 'list')
        self.assertTrue(a == b)
        # test with labs: read
        c = self.ac.rest('labs', 'read', 'mainenlab')
        self.assertTrue([lab for lab in a if
                         lab['name'] == 'mainenlab'][0] == c)
        # test read with UUID object
        dset = self.ac.rest('datasets', 'read', id=UUID('738eca6f-d437-40d6-a9b8-a3f4cbbfbff7'))
        self.assertEqual(dset['name'], '_iblrig_videoCodeFiles.raw.zip')
        # Test with full URL
        d = self.ac.rest(
            'labs', 'read',
            f'{TEST_DB_1["base_url"]}/labs/mainenlab')
        self.assertEqual(c, d)
        # test a more complex endpoint with a filter and a selection
        sub = self.ac.rest('subjects/flowers', 'list')
        sub1 = self.ac.rest('subjects?nickname=flowers', 'list')
        self.assertTrue(len(sub1) == 1)
        self.assertEqual(sub['nickname'], sub1[0]['nickname'])
        # also make sure the action is overriden on a filter query
        sub2 = self.ac.rest('/subjects?nickname=flowers')
        self.assertEqual(sub1, sub2)

    def test_download_datasets_with_api(self):
        ac_public = wc.AlyxClient(**TEST_DB_2)
        cache_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(cache_dir))

        # Test 1: empty dir, dict mode
        dset = ac_public.get('/datasets/' + self.test_data_uuid)
        urls = wc.dataset_record_to_url(dset)
        url = [u for u in urls if u.startswith('https://ibl.flatiron')]
        file_name, = ac_public.download_file(url, target_dir=cache_dir)
        self.assertTrue(os.path.isfile(file_name))
        os.unlink(file_name)

        # Test 2: empty dir, list mode
        dset = ac_public.get('/datasets?id=' + self.test_data_uuid)
        urls = wc.dataset_record_to_url(dset)
        url = [u for u in urls if u.startswith('https://ibl.flatiron')]
        file_name, = ac_public.download_file(url, target_dir=cache_dir)
        self.assertTrue(os.path.isfile(file_name))
        os.unlink(file_name)

        # Test 3: Log unauthorized error with url (using test alyx)
        url = next(x['data_url'] for x in ac.get('/datasets?exists=True')[0]['file_records'])
        old_par = ac._par
        ac._par = ac._par.set('HTTP_DATA_SERVER_PWD', 'foobar')
        with self.assertLogs(logging.getLogger('one.webclient'), logging.ERROR) as log:
            raised = False
            try:
                ac.download_file(url, target_dir=cache_dir)
                self.assertTrue(url in log.output[-1])
            except Exception as ex:
                # Check error message mentions the HTTP_DATA_SERVER params
                self.assertTrue('HTTP_DATA_SERVER_PWD' in str(ex))
                raised = True
            finally:
                self.assertTrue(raised)
                ac._par = old_par

    def test_download_datasets(self):
        # test downloading a single file
        full_link_to_file = (
            'https://ibl.flatironinstitute.org/public/mrsicflogellab/Subjects/SWC_038/'
            '2020-07-29/001/alf/probes.description.f67570ac-1e54-4ce1-be5d-de2017a42116.json'
        )
        file_name, md5 = wc.http_download_file(full_link_to_file,
                                               return_md5=True, clobber=True)
        with open(file_name, 'r') as json_file:
            data = json.load(json_file)
        self.assertTrue(len(data) > 0)
        self.assertTrue(hashfile.md5(file_name) == md5)

        self.assertFalse(wc.http_download_file('', clobber=True))

        # test downloading a list of files
        links = [full_link_to_file,
                 r'https://ibl.flatironinstitute.org/public/hoferlab/Subjects/SWC_043/'
                 r'2020-09-21/001/alf/probes.description.c4df1eea-c92c-479f-a907-41fa6e770094.json'
                 ]
        file_list = wc.http_download_file_list(links, username=par.HTTP_DATA_SERVER_LOGIN,
                                               password=par.HTTP_DATA_SERVER_PWD)
        for file in file_list:
            with open(file, 'r') as json_file:
                data = json.load(json_file)
            self.assertTrue(len(data) > 0)

    def test_rest_all_actions(self):
        """Test for AlyxClient.rest method using subjects endpoint"""
        # randint reduces conflicts with parallel tests
        nickname = f'foobar_{random.randint(0, 10000)}'
        newsub = {
            'nickname': nickname,
            'responsible_user': 'olivier',
            'birth_date': '2019-06-15',
            'death_date': None,
            'lab': 'cortexlab',
        }
        # look for the subject, create it if necessary
        sub = self.ac.get(f'/subjects?&nickname={nickname}', expires=True)
        if sub:
            self.ac.rest('subjects', 'delete', id=nickname)
        self.ac.rest('subjects', 'create', data=newsub)
        # partial update and full update
        newsub = self.ac.rest('subjects', 'partial_update',
                              id=nickname, data={'description': 'hey'})
        self.assertEqual(newsub['description'], 'hey')
        newsub['description'] = 'hoy'
        newsub = self.ac.rest('subjects', 'update', id=nickname, data=newsub)
        self.assertEqual(newsub['description'], 'hoy')
        # read
        newsub_ = self.ac.rest('subjects', 'read', id=nickname)
        self.assertEqual(newsub, newsub_)
        # list with filter
        sub = self.ac.rest('subjects', 'list', nickname=nickname)
        self.assertEqual(sub[0]['nickname'], newsub['nickname'])
        self.assertTrue(len(sub) == 1)
        # delete
        self.ac.rest('subjects', 'delete', id=nickname)
        self.ac.clear_rest_cache()  # Make sure we hit db
        sub = self.ac.get(f'/subjects?&nickname={nickname}', expires=True)
        self.assertFalse(sub)

    def test_endpoints_docs(self):
        """Test for AlyxClient.list_endpoints method and AlyxClient.rest"""
        # Test endpoint documentation and validation
        endpoints = self.ac.list_endpoints()
        self.assertTrue('auth-token' not in endpoints)
        # Check that calling rest method with no args prints endpoints
        with unittest.mock.patch('sys.stdout', new_callable=io.StringIO) as stdout:
            self.ac.rest()
            self.assertTrue(k in stdout.getvalue() for k in endpoints)
        # Same but with no action
        with unittest.mock.patch('sys.stdout', new_callable=io.StringIO) as stdout:
            self.ac.rest('sessions')
            actions = self.ac.rest_schemes['sessions'].keys()
            self.assertTrue(all(k in stdout.getvalue() for k in actions))
        # Check logs warning when no id provided
        with self.assertLogs(logging.getLogger('one.webclient'), logging.WARNING):
            self.assertIsNone(self.ac.rest('sessions', 'read'))
        # Check logs warning when creating record with missing data
        with self.assertLogs(logging.getLogger('one.webclient'), logging.WARNING):
            self.assertIsNone(self.ac.rest('sessions', 'create'))
        with self.assertRaises(ValueError) as e:
            self.ac.json_field_write('foobar')
        self.assertTrue(k in str(e.exception) for k in endpoints)


class TestMisc(unittest.TestCase):
    def test_update_url_params(self):
        """Test for one.webclient.update_url_params"""
        url = wc.update_url_params('website.com/?q=', {'pg': 5})
        self.assertEqual('website.com/?pg=5', url)

        # Check handles lists
        url = wc.update_url_params('website.com?q=xxx', {'pg': 5, 'foo': ['bar', 'baz']})
        self.assertEqual('website.com?q=xxx&pg=5&foo=bar&foo=baz', url)

        # Check encodes special chars; handles partial URL
        url = '/path?param1=foo bar'
        new_url = wc.update_url_params(url, {'param2': '#2020-01-03#,#2021-02-01#'})
        expected = '/path?param1=foo+bar&param2=%232020-01-03%23%2C%232021-02-01%23'
        self.assertEqual(expected, new_url)

    def test_validate_file_url(self):
        """Test for AlyxClient._validate_file_url"""
        # Should assert that domain matches data server parameter
        with self.assertRaises(AssertionError):
            ac._validate_file_url('https://webserver.net/path/to/file')
        # Should check that the domain is equal and return same URL
        expected = ac._par.HTTP_DATA_SERVER + '/path/to/file.ext'
        self.assertEqual(ac._validate_file_url(expected), expected)
        # Should prepend data server URL
        self.assertEqual(ac._validate_file_url('/path/to/file.ext'), expected)

    def test_no_cache_context_manager(self):
        """Test for one.webclient.no_cache function"""
        assert ac.cache_mode is not None
        with wc.no_cache(ac):
            self.assertIsNone(ac.cache_mode)
        self.assertIsNotNone(ac.cache_mode)


if __name__ == '__main__':
    unittest.main(exit=False, verbosity=2)



"""A module for inter-converting experiment identifiers.

There are multiple ways to uniquely identify an experiment:
    - eid (str) : An experiment UUID as a string
    - np (int64) : An experiment UUID encoded as 2 int64s
    - path (Path) : A pathlib ALF path of the form `<lab>/Subjects/<subject>/<date>/<number>`
    - ref (str) : An experiment reference string of the form `yyyy-mm-dd_n_subject`
    - url (str) : A remote http session path of the form `<lab>/Subjects/<subject>/<date>/<number>`
"""
import re
import functools
import datetime
from uuid import UUID
from inspect import unwrap
from pathlib import Path, PurePosixPath
from urllib.parse import urlsplit
from typing import Optional, Union, Mapping, List, Iterable as Iter

import numpy as np
import pandas as pd
from iblutil.io import parquet
from iblutil.util import Bunch

import one.alf.io as alfio
from one.alf.spec import is_session_path, is_uuid_string
from one.alf.files import get_session_path, add_uuid_string, session_path_parts
from .util import Listable, ensure_list


def recurse(func):
    """Decorator to call decorated function recursively if first arg is non-string iterable.

    Allows decorated methods to accept both single values, and lists/tuples of values.  When
    given the latter, a list is returned.  This decorator is intended to work on class methods,
    therefore the first arg is assumed to be the object.  Maps and pandas objects are not
    iterated over.

    Parameters
    ----------
    func : function
        A method to decorate

    Returns
    -------
    function
        The decorated method
    """

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        if len(args) <= 1:
            return func(*args, **kwargs)
        obj, first = args[:2]
        exclude = (str, Mapping, pd.Series, pd.DataFrame)
        if isinstance(first, Iter) and not isinstance(first, exclude):
            return [func(obj, item, *args[2:], **kwargs) for item in first]
        else:
            return func(obj, first, *args[2:], **kwargs)

    return wrapper_decorator


def parse_values(func):
    """Convert str values in reference dict to appropriate type.

    Example
    -------
    >>> parse_values(lambda x: x)({'date': '2020-01-01', 'sequence': '001'}, parse=True)
    {'date': datetime.date(2020, 1, 1), 'sequence': 1}
    """

    def parse_ref(ref):
        if ref:
            if isinstance(ref["date"], str):
                if len(ref["date"]) == 10:
                    ref["date"] = datetime.date.fromisoformat(ref["date"])
                else:
                    ref["date"] = datetime.datetime.fromisoformat(ref["date"]).date()
            ref["sequence"] = int(ref["sequence"])
        return ref

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        parse = kwargs.pop("parse", True)
        ref = func(*args, **kwargs)
        if not parse or isinstance(ref, str):
            return ref
        elif isinstance(ref, list):
            return list(map(parse_ref, ref))
        else:
            return parse_ref(ref)

    return wrapper_decorator


class ConversionMixin:
    """A mixin providing methods to interconvert experiment identifiers"""

    def __init__(self):
        self._cache = None
        self._par = None

    @recurse
    def to_eid(
        self,
        id: Listable(Union[str, Path, UUID, dict]) = None,
        cache_dir: Optional[Union[str, Path]] = None,
    ) -> Listable(str):
        """Given any kind of experiment identifier, return a corresponding eid string.

        NB: Currently does not support integer IDs.

        Parameters
        ----------
        id : str, pathlib.Path, UUID, dict, tuple, list
            An experiment identifier
        cache_dir : pathlib.Path, str
            An optional cache directory path for intermittent conversion to path

        Returns
        -------
        str, None
            An experiment ID string or None if session not in cache

        Raises
        ------
        ValueError
            Input ID invalid
        """
        # TODO Could add np2str here
        # if isinstance(id, (list, tuple)):  # Recurse
        #     return [self.to_eid(i, cache_dir) for i in id]
        if id is None:
            return
        elif isinstance(id, UUID):
            return str(id)
        elif self.is_exp_ref(id):
            return self.ref2eid(id)
        elif isinstance(id, dict):
            assert {"subject", "number", "lab"}.issubset(id)
            root = Path(cache_dir or self.cache_dir)
            id = root.joinpath(
                id["lab"],
                "Subjects",
                id["subject"],
                str(id.get("date") or id["start_time"][:10]),
                ("%03d" % id["number"]),
            )

        if isinstance(id, Path):
            return self.path2eid(id)
        elif isinstance(id, str):
            if is_session_path(id) or get_session_path(id):
                return self.path2eid(id)
            if len(id) > 36:
                id = id[-36:]
            if not is_uuid_string(id):
                raise ValueError("Invalid experiment ID")
            else:
                return id
        else:
            raise ValueError("Unrecognized experiment ID")

    @recurse
    def eid2path(self, eid: str) -> Optional[Listable(Path)]:
        """
        From an experiment id or a list of experiment ids, gets the local cache path

        Parameters
        ----------
        eid : str, uuid.UUID
            Experiment ID (UUID) or list of UUIDs

        Returns
        -------
        pathlib.Path
            A session path
        """
        # If not valid return None
        if not is_uuid_string(eid):
            raise ValueError(eid + " is not a valid eID/UUID string")
        if self._cache["sessions"].size == 0:
            return

        # load path from cache
        if self._index_type() is int:
            eid = parquet.str2np(eid).tolist()
        try:
            ses = self._cache["sessions"].loc[eid].squeeze()
            assert isinstance(ses, pd.Series), "Duplicate eids in sessions table"
            return session_record2path(ses.to_dict(), self.cache_dir)
        except KeyError:
            return

    @recurse
    def path2eid(self, path_obj):
        """
        From a local path, gets the experiment id

        Parameters
        ----------
        path_obj : pathlib.Path, str
            Local path or list of local paths

        Returns
        -------
        eid, list
            Experiment ID (eid) string or list of eids
        """
        # else ensure the path ends with mouse,date, number
        session_path = get_session_path(path_obj)
        sessions = self._cache["sessions"]

        # if path does not have a date and a number, or cache is empty return None
        if session_path is None or sessions.size == 0:
            return None

        # reduce session records from cache
        toDate = datetime.date.fromisoformat
        subject, date, number = session_path.parts[-3:]
        for col, val in zip(("subject", "date", "number"), (subject, toDate(date), int(number))):
            sessions = sessions[sessions[col] == val]
            if sessions.size == 0:
                return

        assert len(sessions) == 1

        (eid,) = sessions.index.values
        if isinstance(eid, tuple):
            eid = parquet.np2str(np.array(eid))
        return eid

    @recurse
    def path2record(self, path) -> pd.Series:
        """Convert a file or session path to a dataset or session cache record.

        NB: Assumes <lab>/Subjects/<subject>/<date>/<number> pattern

        Parameters
        ----------
        path : str, pathlib.Path
            Local path or HTTP URL

        Returns
        -------
        pandas.Series
            A cache file record
        """
        is_session = is_session_path(path)
        rec = self._cache["sessions" if is_session else "datasets"]
        if rec.empty:
            return
        # if (rec := self._cache['datasets']).empty:  # py 3.8
        #     return

        if is_session_path(path):
            lab, subject, date, number = session_path_parts(path)
            rec = rec[
                (rec["lab"] == lab)
                & (rec["subject"] == subject)
                & (rec["number"] == int(number))
                & (rec["date"] == datetime.date.fromisoformat(date))
            ]
            return None if rec.empty else rec.squeeze()

        # Deal with file path
        if isinstance(path, str) and path.startswith("http"):
            # Remove the UUID from path
            path = urlsplit(path).path.strip("/")
            path = alfio.remove_uuid_file(PurePosixPath(path), dry=True)
            session_path = get_session_path(path).as_posix()
        else:
            # No way of knowing root session path parts without cache tables
            eid = self.path2eid(path)
            session_series = self.list_datasets(eid, details=True).session_path
            if not eid or session_series.empty:
                return
            session_path, *_ = session_series

        rec = rec[rec["session_path"] == session_path]
        rec = rec[rec["rel_path"].apply(lambda x: path.as_posix().endswith(x))]
        assert len(rec) < 2, "Multiple records found"
        return None if rec.empty else rec.squeeze()

    @recurse
    def path2url(self, filepath):
        """
        Given a local file path, constructs the URL of the remote file.

        Parameters
        ----------
        filepath : str, pathlib.Path
            A local file path

        Returns
        -------
        str
            A remote URL string
        """
        record = self.path2record(filepath)
        if record is None:
            return
        return unwrap(self.record2url)(record)

    def record2url(self, record):
        """Convert a session or dataset record to a remote URL

        NB: Requires online instance

        Parameters
        ----------
        record : pd.Series, pd.DataFrame
            A datasets or sessions cache record.  If DataFrame, iterate over and returns list.

        Returns
        -------
        str, list
            A dataset URL or list if input is DataFrame
        """
        webclient = getattr(self, "_web_client", False)
        assert webclient, "No Web client found for instance"
        # FIXME Should be OneAlyx converter only
        if isinstance(record, pd.DataFrame):
            return [self.record2url(r) for _, r in record.iterrows()]
        if isinstance(record, pd.Series):
            is_session_record = "rel_path" not in record
            if is_session_record:
                # NB: This assumes the root path is in the webclient URL
                session_spec = "{lab}/Subjects/{subject}/{date}/{number:03d}"
                url = record.get("session_path") or session_spec.format(**record)
                return webclient.rel_path2url(url)
            if all(isinstance(x, (int, np.int64)) for x in record.name):
                (uuid,) = parquet.np2str(np.array([record.name[-2:]]))
            else:
                uuid = ensure_list(record.name)[-1]  # may be (eid, did) or simply did
        else:
            raise ValueError("Could not determine uuid")

        session_path, rel_path = record[["session_path", "rel_path"]].to_numpy().flatten()
        url = PurePosixPath(session_path, rel_path)
        return webclient.rel_path2url(add_uuid_string(url, uuid).as_posix())

    def record2path(self, dataset) -> Optional[Path]:
        """
        Given a set of dataset records, checks the corresponding exists flag in the cache
        correctly reflects the files system

        Parameters
        ----------
        dataset : pd.DataFrame, pd.Series
            A datasets dataframe slice

        Returns
        -------
        pathlib.Path
            File path for the record
        """
        assert isinstance(dataset, pd.Series) or len(dataset) == 1
        session_path, rel_path = dataset[["session_path", "rel_path"]].to_numpy().flatten()
        file = Path(self.cache_dir, session_path, rel_path)
        return file  # files[0] if len(datasets) == 1 else files

    @recurse
    def eid2ref(self, eid: Union[str, Iter], as_dict=True, parse=True) -> Union[str, Mapping, List]:
        """
        Get human-readable session ref from path

        Parameters
        ----------
        eid : str, uuid.UUID
            The experiment uuid to find reference for
        as_dict : bool
            If false a string is returned in the form 'subject_sequence_yyyy-mm-dd'
        parse : bool
            If true, the reference date and sequence are parsed from strings to their respective
            data types

        Returns
        -------
        dict, str, list
            One or more objects with keys ('subject', 'date', 'sequence'), or strings with the
            form yyyy-mm-dd_n_subject

        Examples
        --------
        >>> eid = '4e0b3320-47b7-416e-b842-c34dc9004cf8'
        >>> one.eid2ref(eid)
        {'subject': 'flowers', 'date': datetime.date(2018, 7, 13), 'sequence': 1}
        >>> one.eid2ref(eid, parse=False)
        {'subject': 'flowers', 'date': '2018-07-13', 'sequence': '001'}
        >>> one.eid2ref(eid, as_dict=False)
        '2018-07-13_1_flowers'
        >>> one.eid2ref(eid, as_dict=False, parse=False)
        '2018-07-13_001_flowers'
        >>> one.eid2ref([eid, '7dc3c44b-225f-4083-be3d-07b8562885f4'])
        [{'subject': 'flowers', 'date': datetime.date(2018, 7, 13), 'sequence': 1},
         {'subject': 'KS005', 'date': datetime.date(2019, 4, 11), 'sequence': 1}]
        """
        d = self.get_details(eid)
        if parse:
            ref = {"subject": d["subject"], "date": d["date"], "sequence": d["number"]}
            format_str = "{date:%Y-%m-%d}_{sequence:d}_{subject:s}"
        else:
            ref = {
                "subject": d["subject"],
                "date": str(d["date"]),
                "sequence": "%03d" % d["number"],
            }
            format_str = "{date:s}_{sequence:s}_{subject:s}"
        return Bunch(ref) if as_dict else format_str.format(**ref)

    @recurse
    def ref2eid(self, ref: Union[Mapping, str, Iter]) -> Union[str, List]:
        """
        Returns experiment uuid, given one or more experiment references

        Parameters
        ----------
        ref : str, dict, list
            One or more objects with keys ('subject', 'date', 'sequence'), or strings with
            the form yyyy-mm-dd_n_subject

        Returns
        -------
        str, list
            One or more experiment uuid strings

        Examples
        --------
        >>> base = 'https://test.alyx.internationalbrainlab.org'
        >>> one = ONE(username='test_user', password='TapetesBloc18', base_url=base)
        Connected to...
        >>> ref = {'date': datetime(2018, 7, 13).date(), 'sequence': 1, 'subject': 'flowers'}
        >>> one.ref2eid(ref)
        '4e0b3320-47b7-416e-b842-c34dc9004cf8'
        >>> one.ref2eid(['2018-07-13_1_flowers', '2019-04-11_1_KS005'])
        ['4e0b3320-47b7-416e-b842-c34dc9004cf8',
         '7dc3c44b-225f-4083-be3d-07b8562885f4']
        """
        ref = self.ref2dict(ref, parse=False)  # Ensure dict
        session = self.search(subject=ref["subject"], date_range=str(ref["date"]), number=ref["sequence"])
        assert len(session) == 1, "session not found"
        return session[0]

    @recurse
    def ref2path(self, ref):
        """
        Convert one or more experiment references to session path(s)

        Parameters
        ----------
        ref : str, dict, list
            One or more objects with keys ('subject', 'date', 'sequence'), or strings with
            the form yyyy-mm-dd_n_subject

        Returns
        -------
        pathlib.Path
            Path object(s) for the experiment session(s)

        Examples
        --------
        >>> base = 'https://test.alyx.internationalbrainlab.org'
        >>> one = ONE(username='test_user', password='TapetesBloc18', base_url=base)
        Connected to...
        >>> ref = {'subject': 'flowers', 'date': datetime(2018, 7, 13).date(), 'sequence': 1}
        >>> one.ref2path(ref)
        WindowsPath('E:/FlatIron/zadorlab/Subjects/flowers/2018-07-13/001')
        >>> one.ref2path(['2018-07-13_1_flowers', '2019-04-11_1_KS005'])
        [WindowsPath('E:/FlatIron/zadorlab/Subjects/flowers/2018-07-13/001'),
         WindowsPath('E:/FlatIron/cortexlab/Subjects/KS005/2019-04-11/001')]
        """
        eid2path = unwrap(self.eid2path)
        ref2eid = unwrap(self.ref2eid)
        return eid2path(self, ref2eid(self, ref))

    @staticmethod
    @parse_values
    def path2ref(path_str: Union[str, Path, Iter], as_dict=True) -> Union[Bunch, List]:
        """
        Returns a human readable experiment reference, given a session path.
        The path need not exist.

        Parameters
        ----------
        path_str : str
            A path to a given session
        as_dict : bool
            If True a Bunch is returned, otherwise a string

        Returns
        -------
        dict, str, list
            One or more objects with keys ('subject', 'date', 'sequence')

        Examples
        --------
        >>> path_str = Path('E:/FlatIron/Subjects/zadorlab/flowers/2018-07-13/001')
        >>> path2ref(path_str)
        {'subject': 'flowers', 'date': datetime.date(2018, 7, 13), 'sequence': 1}
        >>> path2ref(path_str, parse=False)
        {'subject': 'flowers', 'date': '2018-07-13', 'sequence': '001'}
        >>> path_str2 = Path('E:/FlatIron/Subjects/churchlandlab/CSHL046/2020-06-20/002')
        >>> path2ref([path_str, path_str2])
        [{'subject': 'flowers', 'date': datetime.date(2018, 7, 13), 'sequence': 1},
         {'subject': 'CSHL046', 'date': datetime.date(2020, 6, 20), 'sequence': 2}]
        """
        if isinstance(path_str, (list, tuple)):
            return [unwrap(ConversionMixin.path2ref)(x) for x in path_str]
        pattern = r"(?P<subject>[\w-]+)([\\/])(?P<date>\d{4}-\d{2}-\d{2})(\2)(?P<sequence>\d{3})"
        match = re.search(pattern, str(path_str))
        if match:
            ref = match.groupdict()
            return Bunch(ref) if as_dict else "{date:s}_{sequence:s}_{subject:s}".format(**ref)

    @staticmethod
    def is_exp_ref(ref: Union[str, Mapping, Iter]) -> Union[bool, List[bool]]:
        """
        Returns True is ref is a valid experiment reference

        Parameters
        ----------
        ref : str, dict, list
            One or more objects with keys ('subject', 'date', 'sequence'), or strings with
            the form yyyy-mm-dd_n_subject

        Returns
        -------
        bool
            True if ref is valid

        Examples
        --------
        >>> ref = {'date': datetime(2018, 7, 13).date(), 'sequence': 1, 'subject': 'flowers'}
        >>> is_exp_ref(ref)
        True
        >>> is_exp_ref('2018-07-13_001_flowers')
        True
        >>> is_exp_ref('invalid_ref')
        False
        """
        if isinstance(ref, (list, tuple)):
            return [ConversionMixin.is_exp_ref(x) for x in ref]
        if isinstance(ref, (Bunch, dict)):
            if not {"subject", "date", "sequence"}.issubset(ref):
                return False
            ref = "{date}_{sequence}_{subject}".format(**ref)
        elif not isinstance(ref, str):
            return False
        return re.compile(r"\d{4}(-\d{2}){2}_(\d{1}|\d{3})_\w+").match(ref) is not None

    @staticmethod
    def path2refpath(path: str) -> str:
        ALYX_PATH_EID = r"\w+(?:\\|\/)\d{4}\-\d{2}\-\d{2}(?:\\|\/)\d{3}"
        return re.findall(ALYX_PATH_EID, path)[0]

    @staticmethod
    @parse_values
    def ref2dict(ref: Union[str, Mapping, Iter]) -> Union[Bunch, List]:
        """
        Returns a Bunch (dict-like) from a reference string (or list thereof)

        Parameters
        ----------
        ref : str, list
            One or more experiment reference strings

        Returns
        -------
        iblutil.util.Bunch
            A Bunch in with keys ('subject', 'sequence', 'date')

        Examples
        --------
        >>> ref2dict('2018-07-13_1_flowers')
        {'date': datetime.date(2018, 7, 13), 'sequence': 1, 'subject': 'flowers'}
        >>> ref2dict('2018-07-13_001_flowers', parse=False)
        {'date': '2018-07-13', 'sequence': '001', 'subject': 'flowers'}
        >>> ref2dict(['2018-07-13_1_flowers', '2020-01-23_002_ibl_witten_01'])
        [{'date': datetime.date(2018, 7, 13), 'sequence': 1, 'subject': 'flowers'},
         {'date': datetime.date(2020, 1, 23), 'sequence': 2, 'subject': 'ibl_witten_01'}]
        """
        if isinstance(ref, (list, tuple)):
            return [ConversionMixin.ref2dict(x) for x in ref]
        if isinstance(ref, (Bunch, dict)):
            return Bunch(ref)  # Short circuit
        ref = dict(zip(["date", "sequence", "subject"], ref.split("_", 2)))
        return Bunch(ref)

    @staticmethod
    def dict2ref(ref_dict) -> Union[str, List]:
        """
        Convert an experiment reference dict to a string in the format yyyy-mm-dd_n_subject.

        Parameters
        ----------
        ref_dict : dict, Bunch, list, tuple
            A map with the keys ('subject', 'date', 'sequence')

        Returns
        -------
        str, list:
            An experiment reference string, or list thereof
        """
        if isinstance(ref_dict, (list, tuple)):
            return [ConversionMixin.dict2ref(x) for x in ref_dict]
        if not ref_dict:
            return
        parsed = any(not isinstance(k, str) for k in ref_dict.values())
        format_str = "{date:%Y-%m-%d}_{sequence:d}_{subject:s}" if parsed else "{date:s}_{sequence:s}_{subject:s}"
        return format_str.format(**ref_dict)


def one_path_from_dataset(dset, one_cache):
    """
    Returns local one file path from a dset record or a list of dsets records from REST.
    Unlike `to_eid`, this function does not require ONE, and the dataset may not exist.

    Parameters
    ----------
    dset : dict, list
        Dataset dictionary or list of dictionaries from Alyx rest endpoint
    one_cache : str, pathlib.Path, pathlib.PurePath
        The local ONE data cache directory

    Returns
    -------
    pathlib.Path
        The local path for a given dataset
    """
    return path_from_dataset(dset, root_path=one_cache, uuid=False)


def path_from_dataset(dset, root_path=PurePosixPath("/"), repository=None, uuid=False):
    """
    Returns the local file path from a dset record from a REST query.
    Unlike `to_eid`, this function does not require ONE, and the dataset may not exist.

    Parameters
    ----------
    dset : dict, list
        Dataset dictionary or list of dictionaries from Alyx rest endpoint
    root_path : str, pathlib.Path, pathlib.PurePath
        The prefix path such as the ONE download directory or remote http server root
    repository : str, None
        Which data repository to use from the file_records list, defaults to first online
        repository
    uuid : bool
        If True, the file path will contain the dataset UUID

    Returns
    -------
    pathlib.Path, list
        File path or list of paths
    """
    if isinstance(dset, list):
        return [path_from_dataset(d) for d in dset]
    if repository:
        fr = next((fr for fr in dset["file_records"] if fr["data_repository"] == repository))
    else:
        fr = next((fr for fr in dset["file_records"] if fr["data_url"]))
    uuid = dset["url"][-36:] if uuid else None
    return path_from_filerecord(fr, root_path=root_path, uuid=uuid)


def path_from_filerecord(fr, root_path=PurePosixPath("/"), uuid=None):
    """
    Returns a data file Path constructed from an Alyx file record.  The Path type returned
    depends on the type of root_path: If root_path is a string a Path object is returned,
    otherwise if the root_path is a PurePath, the same path type is returned.

    Parameters
    ----------
    fr : dict
        An Alyx file record dict
    root_path : str, pathlib.Path
        An optional root path
    uuid : str, uuid.UUID
        An optional dataset UUID to add to the file name

    Returns
    -------
    pathlib.Path
        A filepath as a pathlib object
    """
    if isinstance(fr, list):
        return [path_from_filerecord(f) for f in fr]
    repo_path = fr["data_repository_path"]
    repo_path = repo_path[repo_path.startswith("/") :]  # remove starting / if any
    # repo_path = (p := fr['data_repository_path'])[p[0] == '/':]  # py3.8 Remove slash at start
    file_path = PurePosixPath(repo_path, fr["relative_path"])
    if root_path:
        # NB: By checking for string we won't cast any PurePaths
        if isinstance(root_path, str):
            root_path = Path(root_path)
        file_path = root_path / file_path
    if uuid:
        file_path = add_uuid_string(file_path, uuid)
    return file_path


def session_record2path(session, root_dir=None):
    """
    Convert a session record into a path.

    If a lab key is present, the path will be in the form
    root_dir/lab/Subjects/subject/yyyy-mm-dd/nnn, otherwise root_dir/subject/yyyy-mm-dd/nnn.

    Parameters
    ----------
    session : Mapping
        A session record with keys ('subject', 'date', 'number'[, 'lab']).
    root_dir : str, pathlib.Path, pathlib.PurePath
        A root directory to prepend.

    Returns
    -------
    pathlib.Path, Pathlib.PurePath
        A constructed path of the session.

    Examples
    --------
    >>> session_record2path({'subject': 'ALK01', 'date': '2020-01-01', 'number': 1})
    PurePosixPath('ALK01/2020-01-01/001')

    >>> record = {'date': datetime.datetime.fromisoformat('2020-01-01').date(),
    ...           'number': '001', 'lab': 'foo', 'subject': 'ALK01'}
    >>> session_record2path(record, Path('/home/user'))
    Path('/home/user/foo/Subjects/ALK01/2020-01-01/001')
    """
    rel_path = PurePosixPath(
        session.get("lab") if session.get("lab") else "",
        "Subjects" if session.get("lab") else "",
        session["subject"],
        str(session["date"]),
        str(session["number"]).zfill(3),
    )
    if not root_dir:
        return rel_path
    elif isinstance(root_dir, str):
        root_dir = Path(root_dir)
    return Path(root_dir).joinpath(rel_path)



"""API for interacting with a remote Alyx instance through REST
The AlyxClient class contains methods for making remote Alyx REST queries and downloading remote
files through Alyx.

Examples
--------
>>> alyx = AlyxClient(
...     username='test_user', password='TapetesBloc18',
...     base_url='https://test.alyx.internationalbrainlab.org')

List subjects

>>> subjects = alyx.rest('subjects', 'list')

Create a subject

>>> record = {
...     'nickname': nickname,
...     'responsible_user': 'olivier',
...     'birth_date': '2019-06-15',
...     'death_date': None,
...     'lab': 'cortexlab',
... }
>>> new_subj = alyx.rest('subjects', 'create', data=record)

Download a remote file, given a local path

>>> url = 'zadorlab/Subjects/flowers/2018-07-13/1/channels.probe.npy'
>>> local_path = alyx.download_file(url, target_dir='zadorlab/Subjects/flowers/2018-07-13/1/')
"""

import json
import logging
import math
import os
import re
import functools
import urllib.request
from urllib.error import HTTPError
import urllib.parse
from collections.abc import Mapping
from typing import Optional
from datetime import datetime, timedelta
from pathlib import Path
import warnings
import hashlib
import zipfile
import tempfile
from getpass import getpass
from contextlib import contextmanager

import requests
from tqdm import tqdm

from pprint import pprint
import one.params
from iblutil.io import hashfile
from iblutil.io.params import set_hidden
from one.util import ensure_list
import concurrent.futures

from types import MethodType

_logger = logging.getLogger(__name__)


def _cache_response(method):
    """Decorator for the generic request method

    Caches the result of the query and on subsequent calls, returns cache instead of hitting the
    database

    Parameters
    ----------
    method : function
        Function to wrap (i.e. AlyxClient._generic_request)

    Returns
    -------
    function
        Handle to wrapped method
    """

    @functools.wraps(method)
    def wrapper_decorator(alyx_client, *args, expires=None, clobber=False, **kwargs):
        """
        REST caching wrapper

        Parameters
        ----------
        alyx_client : AlyxClient
            An instance of the AlyxClient class
        args : any
            Positional arguments for applying to wrapped function
        expires : bool
            An optional timedelta for how long cached response is valid.  If True, the cached
            response will not be used on subsequent calls.  If None, the default expiry is applied.
        clobber : bool
            If True any existing cached response is overwritten
        kwargs : any
            Keyword arguments for applying to wrapped function

        Returns
        -------
        dict
            The REST response JSON either from cached file or directly from remote
        """
        expires = expires or alyx_client.default_expiry
        mode = (alyx_client.cache_mode or "").lower()
        if args[0].__name__ != mode and mode != "*":
            return method(alyx_client, *args, **kwargs)
        # Check cache
        rest_cache = alyx_client.cache_dir.joinpath(".rest")
        sha1 = hashlib.sha1()
        sha1.update(bytes(args[1], "utf-8"))
        name = sha1.hexdigest()
        # Reversible but length may exceed 255 chars
        # name = base64.urlsafe_b64encode(args[2].encode('UTF-8')).decode('UTF-8')
        files = list(rest_cache.glob(name))
        cached = None
        if len(files) == 1 and not clobber:
            _logger.debug("loading REST response from cache")
            with open(files[0], "r") as f:
                cached, when = json.load(f)
            if datetime.fromisoformat(when) > datetime.now():
                return cached
        try:
            response = method(alyx_client, *args, **kwargs)
        except requests.exceptions.ConnectionError as ex:
            if cached and not clobber:
                warnings.warn("Failed to connect, returning cached response", RuntimeWarning)
                return cached
            raise ex  # No cache and can't connect to database; re-raise

        # Save response into cache
        if not rest_cache.exists():
            rest_cache.mkdir(parents=True)
            rest_cache = set_hidden(rest_cache, True)

        _logger.debug("caching REST response")
        expiry_datetime = datetime.now() + (timedelta() if expires is True else expires)
        with open(rest_cache / name, "w") as f:
            json.dump((response, expiry_datetime.isoformat()), f)
        return response

    return wrapper_decorator


@contextmanager
def no_cache(ac=None):
    """Temporarily turn off the REST cache for a given Alyx instance.

    This function is particularly useful when calling ONE methods in remote mode.

    Parameters
    ----------
    ac : AlyxClient
        An instance of the AlyxClient to modify.  If None, the a new object is instantiated

    Returns
    -------
    AlyxClient
        The instance of Alyx with cache disabled

    Examples
    --------
    >>> from one.api import ONE
    >>> with no_cache(ONE().alyx):
    ...     eids = ONE().search(subject='foobar', query_type='remote')
    """
    ac = ac or AlyxClient()
    cache_mode = ac.cache_mode
    ac.cache_mode = None
    try:
        yield ac
    finally:
        ac.cache_mode = cache_mode


class _PaginatedResponse(Mapping):
    """
    This class allows to emulate a list from a paginated response.
    Provides cache functionality.

    Examples
    --------
    >>> r = _PaginatedResponse(client, response)
    """

    def __init__(self, alyx, rep, cache_args=None, callbacks=[]):
        """
        A paginated response cache object

        Parameters
        ----------
        alyx : AlyxClient
            An instance of an AlyxClient associated with the REST response
        rep : dict
            A paginated REST response JSON dictionary
        cache_args : dict
            A dict of kwargs to pass to _cache_response decorator upon subsequent requests
        """

        self.alyx = alyx
        self.base_url = self.alyx.base_url
        self.count = rep["count"]
        self.limit = len(rep["results"])
        self._cache_args = cache_args or {}
        # store URL without pagination query params
        self.query = rep["next"]
        # init the cache, list with None with count size
        self._cache = [None] * self.count

        if not isinstance(callbacks, list):
            callbacks = [callbacks]

        self.finishing_callbacks = callbacks

        # fill the cache with results of the query
        self.store_results(rep, 0)

    def __len__(self):
        return self.count

    def __getitem__(self, item):
        """Returns an item from cache if the indices locations are already loaded,
        or populates the cache chunk by chunk if some locations are empty.
        """
        if isinstance(item, slice):
            while None in self._cache[item]:
                # .index(None) finds the first location where the cache is none,
                # and populate stores a new chunk starting from here
                self.populate(self._cache[item].index(None))
        elif item > self.count:
            raise IndexError(
                f"The paginated response has no item at position {item} : " f"it contains only {self.count} items"
            )
        elif self._cache[item] is None:
            self.populate(item)
        return self._cache[item]

    def populate(self, idx):
        """Populate a chunk of size self.limit, from an offset query depending on the value of the index required."""
        offset = self.limit * math.floor(idx / self.limit)
        query = update_url_params(self.query, {"limit": self.limit, "offset": offset})
        res = self.alyx._generic_request(requests.get, query, **self._cache_args)
        if self.count != res["count"]:
            warnings.warn(
                f"remote results for {urllib.parse.urlsplit(query).path} endpoint changed; results may be inconsistent",
                RuntimeWarning,
            )
        self.store_results(res, offset)

    def store_results(self, res, offset):
        for i, r in enumerate(res["results"][: self.count - offset]):
            self._cache[i + offset] = self.finish_result(r)

    def finish_result(self, result):
        """Method that should be overridden in child classes"""
        for callback in self.finishing_callbacks:
            result = callback(result)
        return result

    def add_finishing_callback(self, callback):
        if callback not in self.finishing_callbacks:
            self.finishing_callbacks.append(callback)

    def __iter__(self):
        for i in range(self.count):
            try:
                yield self.__getitem__(i)
            except requests.HTTPError as e:
                _logger.error(
                    f"{e.response.status_code} error while trying to access " f"PaginatedResponse data at position {i}"
                )
                raise e


def update_url_params(url: str, params: dict) -> str:
    """Add/update the query parameters of a URL and make url safe

    Parameters
    ----------
    url : str
        A URL string with which to update the query parameters
    params : dict
        A dict of new parameters.  For multiple values for the same query, use a list (see example)

    Returns
    -------
    str
        A new URL with said parameters updated

    Examples
    -------
    >>> update_url_params('website.com/?q=', {'pg': 5})
    'website.com/?pg=5'

    >>> update_url_params('website.com?q=xxx', {'pg': 5, 'foo': ['bar', 'baz']})
    'website.com?q=xxx&pg=5&foo=bar&foo=baz'
    """
    # Remove percent-encoding
    url = urllib.parse.unquote(url)
    parsed_url = urllib.parse.urlsplit(url)
    # Extract URL query arguments and convert to dict
    parsed_get_args = urllib.parse.parse_qs(parsed_url.query, keep_blank_values=False)
    # Merge URL arguments dict with new params
    parsed_get_args.update(params)
    # Convert back to query string
    encoded_get_args = urllib.parse.urlencode(parsed_get_args, doseq=True)
    # Update parser and convert to full URL str
    return parsed_url._replace(query=encoded_get_args).geturl()


def update_url_listparams(url: str, params: list) -> str:
    # Remove percent-encoding
    url = urllib.parse.unquote(url)
    parsed_url = urllib.parse.urlsplit(url)

    # Extract URL query arguments, merge with existing params
    parsed_get_args = urllib.parse.parse_qsl(parsed_url.query, keep_blank_values=False)
    parsed_get_args += params

    # Convert back to query string
    encoded_get_args = urllib.parse.urlencode(parsed_get_args, doseq=True)
    # Update parser and convert to full URL str
    return parsed_url._replace(query=encoded_get_args).geturl()


def http_download_file_list(links_to_file_list, **kwargs):
    """
    Downloads a list of files from a remote HTTP server from a list of links.
    Generates up to 4 separate threads to handle downloads.
    Same options behaviour as http_download_file.

    Parameters
    ----------
    links_to_file_list : list
        List of http links to files
    kwargs
        Optional arguments to pass to http_download_file

    Returns
    -------
    list of pathlib.Path
        A list of the local full path of the downloaded files.
    """
    links_to_file_list = list(links_to_file_list)  # In case generator was passed
    n_threads = 4  # Max number of threads
    outputs = []
    target_dir = kwargs.pop("target_dir", None)
    # Ensure target dir the length of url list
    if target_dir is None or isinstance(target_dir, (str, Path)):
        target_dir = [target_dir] * len(links_to_file_list)
    assert len(target_dir) == len(links_to_file_list)
    # using with statement to ensure threads are cleaned up promptly
    zipped = zip(links_to_file_list, target_dir)
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_threads) as executor:
        # Multithreading load operations
        futures = [executor.submit(http_download_file, link, target_dir=target, **kwargs) for link, target in zipped]
        zip(links_to_file_list, ensure_list(kwargs.pop("target_dir", None)))
        # TODO Reintroduce variable timeout value based on file size and download speed of 5 Mb/s?
        # timeout = reduce(lambda x, y: x + (y.get('file_size', 0) or 0), dsets, 0) / 625000 ?
        concurrent.futures.wait(futures, timeout=None)
        # build return list
        for future in futures:
            outputs.append(future.result())
    # if returning md5, separate list of tuples into two lists: (files, md5)
    return list(zip(*outputs)) if kwargs.get("return_md5", False) else outputs


def http_download_file(
    full_link_to_file,
    chunks=None,
    *,
    clobber=False,
    silent=False,
    username="",
    password="",
    target_dir="",
    return_md5=False,
    headers=None,
):
    """
    Download a file from a remote HTTP server.

    Parameters
    ----------
    full_link_to_file : str
        HTTP link to the file
    chunks : tuple of ints
        Chunks to download
    clobber : bool
        If True, force overwrite the existing file
    silent : bool
        If True, suppress download progress bar
    username : str
        User authentication for password protected file server
    password : str
        Password authentication for password protected file server
    target_dir : str, pathlib.Path
        Directory in which files are downloaded; defaults to user's Download directory
    return_md5 : bool
        If True an MD5 hash of the file is additionally returned
    headers : list of dicts
        Additional headers to add to the request (auth tokens etc.)

    Returns
    -------
    pathlib.Path
        The full file path of the downloaded file
    """
    if not full_link_to_file:
        return (None, None) if return_md5 else None

    # makes sure special characters get encoded ('#' in file names for example)
    surl = urllib.parse.urlsplit(full_link_to_file, allow_fragments=False)
    full_link_to_file = surl._replace(path=urllib.parse.quote(surl.path)).geturl()

    # default cache directory is the home dir
    if not target_dir:
        target_dir = str(Path.home().joinpath("Downloads"))

    # This is the local file name
    file_name = str(target_dir) + os.sep + os.path.basename(full_link_to_file)

    # do not overwrite an existing file unless specified
    if not clobber and os.path.exists(file_name):
        return (file_name, hashfile.md5(file_name)) if return_md5 else file_name

    # This should be the base url you wanted to access.
    baseurl = os.path.split(str(full_link_to_file))[0]

    # Create a password manager
    manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    if username and password:
        manager.add_password(None, baseurl, username, password)

    # Create an authentication handler using the password manager
    auth = urllib.request.HTTPBasicAuthHandler(manager)

    # Create an opener that will replace the default urlopen method on further calls
    opener = urllib.request.build_opener(auth)
    urllib.request.install_opener(opener)

    # Support for partial download.
    req = urllib.request.Request(full_link_to_file)
    if chunks is not None:
        first_byte, n_bytes = chunks
        req.add_header("Range", "bytes=%d-%d" % (first_byte, first_byte + n_bytes - 1))

    # add additional headers
    if headers is not None:
        for k in headers:
            req.add_header(k, headers[k])

    # Open the url and get the length
    try:
        u = urllib.request.urlopen(req)
    except HTTPError as e:
        _logger.error(f"{str(e)} {full_link_to_file}")
        raise e

    file_size = int(u.getheader("Content-length"))
    if not silent:
        print(f"Downloading: {file_name} Bytes: {file_size}")
    block_sz = 8192 * 64 * 8

    md5 = hashlib.md5()
    f = open(file_name, "wb")
    with tqdm(total=file_size / 1024 / 1024, disable=silent) as pbar:
        while True:
            buffer = u.read(block_sz)
            if not buffer:
                break
            f.write(buffer)
            if return_md5:
                md5.update(buffer)
            pbar.update(len(buffer) / 1024 / 1024)
    f.close()

    return (file_name, md5.hexdigest()) if return_md5 else file_name


def file_record_to_url(file_records) -> list:
    """
    Translate a Json dictionary to an usable http url for downloading files.

    Parameters
    ----------
    file_records : dict
        JSON containing a 'data_url' field

    Returns
    -------
    list of str
        A list of full data urls
    """
    urls = []
    for fr in file_records:
        if fr["data_url"] is not None:
            urls.append(fr["data_url"])
    return urls


def dataset_record_to_url(dataset_record) -> list:
    """
    Extracts a list of files urls from a list of dataset queries.

    Parameters
    ----------
    dataset_record : list, dict
        Dataset JSON from a REST request

    Returns
    -------
    list of str
        A list of file urls corresponding to the datasets records
    """
    urls = []
    if isinstance(dataset_record, dict):
        dataset_record = [dataset_record]
    for ds in dataset_record:
        urls += file_record_to_url(ds["file_records"])
    return urls


class AlyxClient:
    """
    Class that implements simple GET/POST wrappers for the Alyx REST API.
    See https://openalyx.internationalbrainlab.org/docs
    """

    _token = None
    _headers = None  # Headers for REST requests only
    """str: The Alyx username"""
    user = None
    """str: The Alyx database URL"""
    base_url: None | str = None

    def __init__(
        self,
        base_url=None,
        username=None,
        password=None,
        cache_dir=None,
        silent=False,
        cache_rest="GET",
    ):
        """
        Create a client instance that allows to GET and POST to the Alyx server.
        For One, constructor attempts to authenticate with credentials in params.py.
        For standalone cases, AlyxClient(username='', password='', base_url='').

        Parameters
        ----------
        base_url : str
            Alyx server address, including port and protocol
        username : str
            Alyx database user
        password : str
            Alyx database password
        cache_dir : str, pathlib.Path
            The default root download location
        silent : bool
            If true, user prompts and progress bars are suppressed
        cache_rest : str
            Which type of http method to apply cache to; if '*', all requests are cached
        stay_logged_in : bool
            If true, auth token is cached
        """
        self.silent = silent
        self._par = one.params.get(client=base_url, silent=self.silent, username=username)

        self.base_url = base_url or self._par.ALYX_URL
        if self.base_url is None:
            raise ValueError("base_url was None after resolution")

        self._par = self._par.set("CACHE_DIR", cache_dir or self._par.CACHE_DIR)
        if username or password:
            self.authenticate(username, password)
        self._rest_schemes = None
        # the mixed accept application may cause errors sometimes, only necessary for the docs
        self._headers = {**(self._headers or {}), "Accept": "application/json"}
        # REST cache parameters
        # The default length of time that cache file is valid for,
        # The default expiry is overridden by the `expires` kwarg.  If False, the caching is
        # turned off.
        self.default_expiry = timedelta(days=1)
        self.cache_mode = cache_rest
        self._obj_id = id(self)

    @property
    def rest_schemes(self):
        """dict: The REST endpoints and their parameters"""
        # Delayed fetch of rest schemes speeds up instantiation
        if not self._rest_schemes:
            self._rest_schemes = self.get("/docs", expires=timedelta(weeks=1))
            _logger.debug(f"Rest schemes : {self._rest_schemes}")
        return self._rest_schemes

    @property
    def cache_dir(self):
        """pathlib.Path: The location of the downloaded file cache"""
        return Path(self._par.CACHE_DIR)

    def delete_cache(self):
        """Delete all cached files in the .rest directory of your ONE installation "
        "(usually located in ONE inside downloads)"""
        cache_dir = self.cache_dir.joinpath(".rest")
        for item in os.listdir(cache_dir):
            os.remove(os.path.join(cache_dir, item))

    @property
    def is_logged_in(self):
        """bool: Check if user logged into Alyx database; True if user is authenticated"""
        return self._token and self.user and self._headers and "Authorization" in self._headers

    def list_endpoints(self):
        """
        Return a list of available REST endpoints

        Returns
        -------
            List of REST endpoint strings
        """
        EXCLUDE = ("_type", "_meta", "", "auth-token")
        return sorted(x for x in self.rest_schemes.keys() if x not in EXCLUDE)

    @_cache_response
    def _generic_request(self, reqfunction, rest_query, data=None, files=None):
        if not self._token and (not self._headers or "Authorization" not in self._headers):
            self.authenticate(username=self.user)
        # makes sure the base url is the one from the instance

        # make sure we keep only the path and query part of the url, without the netloc and scheme
        rest_query = urllib.parse.urlsplit(rest_query)._replace(netloc="", scheme="").geturl()

        if not rest_query.startswith("/"):
            rest_query = "/" + rest_query
        _logger.debug(f"{self.base_url + rest_query}, headers: {self._headers}")
        headers = self._headers.copy()
        if files is None:
            data = json.dumps(data) if isinstance(data, dict) or isinstance(data, list) else data
            headers["Content-Type"] = "application/json"
        if rest_query.startswith("/docs"):
            # the mixed accept application may cause errors sometimes, only necessary for the docs
            headers["Accept"] = "application/coreapi+json"
        r = reqfunction(
            self.base_url + rest_query,
            stream=True,
            headers=headers,
            data=data,
            files=files,
        )
        if r and r.status_code in (200, 201):
            return json.loads(r.text)
        elif r and r.status_code == 204:
            return
        if r.status_code == 403 and '"Invalid token."' in r.text:
            _logger.debug("Token invalid; Attempting to re-authenticate...")
            # Log out in order to flush stale token.  At this point we no longer have the password
            # but if the user re-instantiates with a password arg it will request a new token.
            username = self.user
            if self.silent:  # no need to log out otherwise; user will be prompted for password
                self.logout()
            self.authenticate(username=username, force=True)
            return self._generic_request(reqfunction, rest_query, data=data, files=files)
        else:
            _logger.debug("Response text: " + r.text)
            try:
                message = json.loads(r.text)
                message.pop("status_code", None)  # Get status code from response object instead
                message = message.get("detail") or message  # Get details if available
            except json.decoder.JSONDecodeError:
                message = r.text
            raise requests.HTTPError(r.status_code, rest_query, message, response=r)

    def authenticate(self, username=None, password=None, cache_token=True, force=False):
        """
        Gets a security token from the Alyx REST API to create requests headers.
        Credentials are loaded via one.params

        Parameters
        ----------
        username : str
            Alyx username.  If None, token not cached and not silent, user is prompted.
        password : str
            Alyx password.  If None, token not cached and not silent, user is prompted.
        cache_token : bool
            If true, the token is cached for subsequent auto-logins
        force : bool
            If true, any cached token is ignored
        """
        # Get username
        if username is None:
            username = getattr(self._par, "ALYX_LOGIN", self.user)
        if username is None and not self.silent:
            username = input("Enter Alyx username:")

        # Check if token cached
        if not force and getattr(self._par, "TOKEN", False) and username in self._par.TOKEN:
            self._token = self._par.TOKEN[username]
            self._headers = {
                "Authorization": f"Token {list(self._token.values())[0]}",
                "Accept": "application/json",
            }
            self.user = username
            return

        # Get password
        if password is None:
            password = getattr(self._par, "ALYX_PWD", None)
        if password is None and not self.silent:
            password = getpass(f'Enter Alyx password for "{username}":')

        try:
            credentials = {"username": username, "password": password}
            rep = requests.post(self.base_url + "/auth-token", data=credentials)
        except requests.exceptions.ConnectionError:
            raise ConnectionError(
                f"Can't connect to {self.base_url}.\n" + "Check your internet connections and Alyx database firewall"
            )
        # Assign token or raise exception on auth error
        if rep.ok:
            self._token = rep.json()
            assert list(self._token.keys()) == ["token"]
        else:
            if rep.status_code == 400:  # Auth error; re-raise with details
                redacted = "*" * len(credentials["password"]) if credentials["password"] else None
                message = (
                    "Alyx authentication failed with credentials: "
                    f"user = {credentials['username']}, password = {redacted}"
                )
                raise requests.HTTPError(rep.status_code, rep.url, message, response=rep)
            else:
                rep.raise_for_status()

        self._headers = {
            "Authorization": "Token {}".format(list(self._token.values())[0]),
            "Accept": "application/json",
        }
        if cache_token:
            # Update saved pars
            par = one.params.get(client=self.base_url, silent=True)
            tokens = getattr(par, "TOKEN", {})
            tokens[username] = self._token
            one.params.save(par.set("TOKEN", tokens), self.base_url)
            # Update current pars
            self._par = self._par.set("TOKEN", tokens)
        self.user = username
        if not self.silent:
            print(f"Connected to {self.base_url} as {self.user}")

    def logout(self):
        """Log out from Alyx
        Deletes the cached authentication token for the currently logged-in user
        """
        if not self.is_logged_in:
            return
        par = one.params.get(client=self.base_url, silent=True)
        username = self.user
        # Remove token from cache
        if getattr(par, "TOKEN", False) and username in par.TOKEN:
            del par.TOKEN[username]
            one.params.save(par, self.base_url)
        # Remove token from local pars
        if getattr(self._par, "TOKEN", False) and username in self._par.TOKEN:
            del self._par.TOKEN[username]
        # Remove token from object
        self.user = None
        self._token = None
        if self._headers and "Authorization" in self._headers:
            del self._headers["Authorization"]
        self.clear_rest_cache()
        if not self.silent:
            print(f"{username} logged out from {self.base_url}")

    def delete(self, rest_query):
        """
        Sends a DELETE request to the Alyx server. Will raise an exception on any status_code
        other than 200, 201.

        Parameters
        ----------
        rest_query : str
            A REST query string either as a relative URL path complete URL

        Returns
        -------
        JSON interpreted dictionary from response

        Examples
        --------
        >>> AlyxClient.delete('/weighings/c617562d-c107-432e-a8ee-682c17f9e698')
        >>> AlyxClient.delete(
        ...     'https://alyx.example.com/endpoint/c617562d-c107-432e-a8ee-682c17f9e698')
        """
        return self._generic_request(requests.delete, rest_query)

    def download_file(self, url, **kwargs):
        """
        Downloads a single file or list of files on the Alyx server from a
        file record REST field URL

        Parameters
        ----------
        url : str, list
            Full url(s) of the file(s)
        kwargs : Any
            WebClient.http_download_file parameters

        Returns
        -------
        Local path(s) of downloaded file(s)
        """
        if isinstance(url, str):
            url = self._validate_file_url(url)
            download_fcn = http_download_file
        else:
            url = (self._validate_file_url(x) for x in url)
            download_fcn = http_download_file_list
        pars = dict(
            silent=kwargs.pop("silent", self.silent),
            target_dir=kwargs.pop("target_dir", self._par.CACHE_DIR),
            username=self._par.HTTP_DATA_SERVER_LOGIN,
            password=self._par.HTTP_DATA_SERVER_PWD,
            **kwargs,
        )
        try:
            files = download_fcn(url, **pars)
        except HTTPError as ex:
            if ex.code == 401:
                ex.msg += (
                    " - please check your HTTP_DATA_SERVER_LOGIN and "
                    "HTTP_DATA_SERVER_PWD ONE params, or username/password kwargs"
                )
            raise ex
        return files

    def download_cache_tables(self, source=None, destination=None):
        """Downloads the Alyx cache tables to the local data cache directory

        Parameters
        ----------
        source : str, pathlib.Path
            The remote HTTP directory of the cache table (excluding the filename).
            Default: AlyxClient.base_url.
        destination : str, pathlib.Path
            The target directory into to which the tables will be downloaded.

        Returns
        -------
            List of parquet table file paths.
        """
        # query the database for the latest cache; expires=None overrides cached response
        self.cache_dir.mkdir(exist_ok=True)
        if not self.is_logged_in:
            self.authenticate()
        source = str(source or f"{self.base_url}/cache.zip")
        destination = destination or self.cache_dir

        headers = self._headers if source.startswith(self.base_url) else None
        with tempfile.TemporaryDirectory(dir=destination) as tmp:
            file = http_download_file(
                source,
                headers=headers,
                silent=self.silent,
                target_dir=tmp,
                clobber=True,
            )
            with zipfile.ZipFile(file, "r") as zipped:
                files = zipped.namelist()
                zipped.extractall(destination)
        return [Path(destination, table) for table in files]

    def _validate_file_url(self, url):
        """Asserts that URL matches HTTP_DATA_SERVER parameter.
        Currently only one remote HTTP server is supported for a given AlyxClient instance.  If
        the URL contains only the relative path part, the full URL is returned.

        Parameters
        ----------
        url : str
            The full or partial URL to validate

        Returns
        -------
            The complete URL

        Examples
        --------
        >>> url = self._validate_file_url('https://webserver.net/path/to/file')
        'https://webserver.net/path/to/file'
        >>> url = self._validate_file_url('path/to/file')
        'https://webserver.net/path/to/file'
        """
        # (timothé) : We don't use Web based file transfert, so i commented this part to avoid assertion errors with
        # admin urls and such
        # if url.startswith('http'):  # A full URL
        #     assert url.startswith(self._par.HTTP_DATA_SERVER), \
        #         ('remote protocol and/or hostname does not match HTTP_DATA_SERVER parameter:\n' +
        #          f'"{url[:40]}..." should start with "{self._par.HTTP_DATA_SERVER}"')
        # elif not url.startswith(self._par.HTTP_DATA_SERVER):
        #     url = self.rel_path2url(url)
        return url

    def rel_path2url(self, path):
        """Given a relative file path, return the remote HTTP server URL.
        It is expected that the remote HTTP server has the same file tree as the local system.

        Parameters
        ----------
        path : str, pathlib.Path
            A relative ALF path (subject/date/number/etc.)

        Returns
        -------
            A URL string
        """
        path = str(path).strip("/")
        assert not path.startswith("http")
        return f"{self._par.HTTP_DATA_SERVER}/{path}"

    # def rel_path2admin_url(self, path):
    #     path = str(path).strip("/")
    #     if path.startswith("http"):
    #         return path
    #     return f"{self._par.ALYX_URL}/{path}"

    # def urlify_dict(self, l_result):
    #     keys_to_update = []
    #     for key, value in l_result.items():
    #         if "admin_url" in key:
    #             keys_to_update.append(key)
    #         if isinstance(value, dict):
    #             l_result[key] = self.urlify_dict(l_result[key])
    #         elif isinstance(value, list):
    #             l_result[key] = self.urlify_list(l_result[key])
    #     for key in keys_to_update:
    #         l_result[key] = self.rel_path2admin_url(l_result[key])

    #     return l_result

    # def urlify_list(self, l_result):
    #     for index, value in enumerate(l_result):
    #         if isinstance(value, dict):
    #             l_result[index] = self.urlify_dict(l_result[index])
    #     return l_result

    # def urlify_paginated_response(self, l_result):
    #     for item in l_result:
    #         if isinstance(item, list):
    #             yield self.urlify_list(item)
    #         elif isinstance(item, dict):
    #             yield self.urlify_dict(item)
    #         else:
    #             raise TypeError

    def fix_url(self, data, fixed_keys=["admin_url"], do_fix=False):

        if isinstance(data, list):
            return [self.fix_url(item, do_fix=False) for item in data]
        elif isinstance(data, dict):
            return {
                key: (self.fix_url(item, do_fix=True) if key in fixed_keys else self.fix_url(item, do_fix=False))
                for key, item in data.items()
            }
        elif isinstance(data, _PaginatedResponse):
            data.add_finishing_callback(self.fix_url)
            return data
        elif do_fix:
            baseurl = urllib.parse.urlsplit(self.base_url)
            return urllib.parse.urlsplit(data)._replace(scheme=baseurl.scheme, netloc=baseurl.netloc).geturl()
        else:
            return data

    # def urlify_result(self, result):
    #     if isinstance(result, _PaginatedResponse):
    #         return self.urlify_paginated_response(result)
    #     elif isinstance(result, dict):
    #         return self.urlify_dict(result)
    #     elif isinstance(result, list):
    #         return self.urlify_list(result)
    #     else:
    #         raise TypeError(f"HTTP Request result was not a dict nor a _PaginatedResponse but type : {type(result)}")

    def get(self, rest_query, **kwargs):
        """
        Sends a GET request to the Alyx server. Will raise an exception on any status_code
        other than 200, 201.
        For the dictionary contents and list of endpoints, refer to:
        https://openalyx.internationalbrainlab.org/docs

        Parameters
        ----------
        rest_query : str
            A REST URL path, e.g. '/sessions?user=Hamish'
        kwargs : any
            Optional arguments to pass to _generic_request and _cache_response decorator

        Returns
        -------
        JSON interpreted dictionary from response
        """
        rep = self._generic_request(requests.get, rest_query, **kwargs)
        if isinstance(rep, dict) and list(rep.keys()) == [
            "count",
            "next",
            "previous",
            "results",
        ]:
            if len(rep["results"]) < rep["count"]:
                cache_args = {k: v for k, v in kwargs.items() if k in ("clobber", "expires")}
                _logger.debug(f"Creating a paginated response for a large Alyx request. Query : {rest_query}")
                rep = _PaginatedResponse(self, rep, cache_args, callbacks=self.fix_url)
                _logger.debug(
                    f"Paginated response total size is : {rep.count}. "
                    f"Limit size is : {rep.limit}. Query is : {rep.query}"
                )
            else:
                rep = rep["results"]
        return self.fix_url(rep)

    def patch(self, rest_query, data=None, files=None):
        """
        Sends a PATCH request to the Alyx server.
        For the dictionary contents, refer to:
        https://openalyx.internationalbrainlab.org/docs

        Parameters
        ----------
        rest_query : str
            The endpoint as full or relative URL
        data : dict, str
            JSON encoded string or dictionary (c.f. requests)
        files : dict, tuple
            Files to attach (c.f. requests)

        Returns
        -------
        Response object
        """
        rep = self._generic_request(requests.patch, rest_query, data=data, files=files)
        return self.fix_url(rep)

    def post(self, rest_query, data=None, files=None):
        """
        Sends a POST request to the Alyx server.
        For the dictionary contents, refer to:
        https://openalyx.internationalbrainlab.org/docs

        Parameters
        ----------
        rest_query : str
            The endpoint as full or relative URL
        data : dict, str
            JSON encoded string or dictionary (c.f. requests)
        files : dict, tuple
            Files to attach (c.f. requests)

        Returns
        -------
        Response object
        """
        rep = self._generic_request(requests.post, rest_query, data=data, files=files)
        return self.fix_url(rep)

    def put(self, rest_query, data=None, files=None):
        """
        Sends a PUT request to the Alyx server.
        For the dictionary contents, refer to:
        https://openalyx.internationalbrainlab.org/docs

        Parameters
        ----------
        rest_query : str
            The endpoint as full or relative URL
        data : dict, str
            JSON encoded string or dictionary (c.f. requests)
        files : dict, tuple
            Files to attach (c.f. requests)

        Returns
        -------
        requests.Response
            Response object
        """
        rep = self._generic_request(requests.put, rest_query, data=data, files=files)
        return self.fix_url(rep)

    def rest(
        self,
        url=None,
        action=None,
        id=None,
        data=None,
        files=None,
        no_cache=False,
        **kwargs,
    ):
        """
        alyx_client.rest(): lists endpoints
        alyx_client.rest(endpoint): lists actions for endpoint
        alyx_client.rest(endpoint, action): lists fields and URL

        Example REST endpoint with all actions:

        >>> client = AlyxClient()
        >>> client.rest('subjects', 'list')
        >>> client.rest('subjects', 'list', field_filter1='filterval')
        >>> client.rest('subjects', 'create', data=sub_dict)
        >>> client.rest('subjects', 'read', id='nickname')
        >>> client.rest('subjects', 'update', id='nickname', data=sub_dict)
        >>> client.rest('subjects', 'partial_update', id='nickname', data=sub_dict)
        >>> client.rest('subjects', 'delete', id='nickname')
        >>> client.rest('notes', 'create', data=nd, files={'image': open(image_file, 'rb')})

        Parameters
        ----------
        url : str
            Endpoint name
        action : str
            One of 'list', 'create', 'read', 'update', 'partial_update', 'delete'
        id : str
            Lookup string for actions 'read', 'update', 'partial_update', and 'delete'
        data : dict
            Data dictionary for actions 'update', 'partial_update' and 'create'
        files : dict, tuple
            Option file(s) to upload
        no_cache : bool
            If true the `list` and `read` actions are performed without returning the cache
        kwargs
            Filters as per the Alyx REST documentation
            cf. https://openalyx.internationalbrainlab.org/docs/

        Returns
        -------
        list, dict
            List of queried dicts ('list') or dict (other actions)
        """

        def get_values_key_chain(dictionary, keys=[]):
            """
            Recursively retrieves all values from a nested dictionary along with their key paths.

            Args:
                dictionary (dict): The dictionary to extract values and key paths from.
                keys (list): The list of keys that leads to the current value. Defaults to an empty list.

            Returns:
                list of tuples: Each tuple contains a value from the dictionary and the list of keys
                                that leads to that value.

            Example:
                Input: dictionary = {'a': 1, 'b': {'c': 2, 'd': {'e': 3}}}
                Output: [(1, ['a']), (2, ['b', 'c']), (3, ['b', 'd', 'e'])]
            """

            # Initialize a list to store value-key path pairs
            pairs = []

            # Iterate over key-value pairs in the current dictionary
            for key, value in dictionary.items():
                # If the value is a dictionary, recursively call the function to handle the nested dict
                if isinstance(value, dict):
                    pairs.extend(get_values_key_chain(value, keys + [key]))
                else:
                    # If the value is not a dictionary, append the value and its key path to the pairs list
                    pairs.append((value, keys + [key]))

            # Return the list of value-key path pairs
            return pairs

        # if endpoint is None, list available endpoints
        if not url:
            pprint(self.list_endpoints())
            return
        # remove beginning slash if any
        if url.startswith("/"):
            url = url[1:]
        # and split to the next slash or question mark
        endpoint = re.findall("^/*[^?/]*", url)[0].replace("/", "")
        # make sure the queried endpoint exists, if not throw an informative error
        if endpoint not in self.rest_schemes.keys():
            av = [k for k in self.rest_schemes.keys() if not k.startswith("_") and k]
            raise ValueError(
                'REST endpoint "'
                + endpoint
                + '" does not exist. Available '
                + "endpoints are \n       "
                + "\n       ".join(av)
            )
        endpoint_scheme = self.rest_schemes[endpoint]
        # on a filter request, override the default action parameter
        if "?" in url:
            action = "list"
        # if action is None, list available actions for the required endpoint
        if not action:
            pprint(list(endpoint_scheme.keys()))
            return
        # make sure the the desired action exists, if not throw an informative error
        if action not in endpoint_scheme:
            raise ValueError(
                'Action "'
                + action
                + '" for REST endpoint "'
                + endpoint
                + '" does '
                + "not exist. Available actions are: "
                + "\n       "
                + "\n       ".join(endpoint_scheme.keys())
            )
        # the actions below require an id in the URL, warn and help the user
        if action in ["read", "update", "partial_update", "delete"] and not id:
            _logger.warning('REST action "' + action + '" requires an ID in the URL: ' + endpoint_scheme[action]["url"])
            return
        # the actions below require a data dictionary, warn and help the user with fields list
        if action in ["create", "update", "partial_update"] and not data:
            pprint(endpoint_scheme[action]["fields"])
            for act in endpoint_scheme[action]["fields"]:
                print("'" + act["name"] + "': ...,")
            _logger.warning('REST action "' + action + '" requires a data dict with above keys')
            return

        # clobber=True means remote request always made, expires=True means response is not cached
        cache_args = {
            "clobber": no_cache,
            "expires": kwargs.pop("expires", False) or no_cache,
        }
        if action == "list":
            # list doesn't require id nor
            assert endpoint_scheme[action]["action"] == "get"
            # add to url data if it is a string
            if id:
                # this is a special case of the list where we query a uuid. Usually read is better
                if "django" in kwargs.keys() and kwargs["django"] != "":
                    kwargs["django"] = kwargs["django"] + ","
                else:
                    kwargs["django"] = ""
                # kwargs["django"] = f"{kwargs['django']}pk,{id}"
                # we remove all other filters from kwargs, as selecting by id is already all or none
                if len(excedent_keys := [key for key in kwargs.keys() if key != "django" and key != "query_type"]):
                    _logger.warning(
                        f"Some fields, {excedent_keys} have been supplied by the user, but an id is present in the list"
                        " search. These fields have been discarded."
                    )
                    # if there is any other filter, we send a warning
                kwargs = {"django": f"{kwargs['django']}pk,{id}"}
            # otherwise, look for a dictionary of filter terms
            if kwargs:
                query_params = []
                for key, value in kwargs.items():
                    # if searched value is a dict, it must be a json compatible field, so we proceed like it.
                    # if not, the backend django server will throw an error anyways
                    if isinstance(value, dict):
                        values_keys = get_values_key_chain(value)
                        values = []
                        for json_value, json_chain_keys in values_keys:

                            # if there is a __lookup_str at the end of the last dict key :
                            lookup_key = "exact"
                            if "__" in json_chain_keys[-1]:
                                lookup_keys = json_chain_keys[-1].split("__")
                                if "not" in lookup_keys:
                                    lookup_keys.pop(lookup_keys.index("not"))

                                if len(lookup_keys) == 1:
                                    lookup_key = "exact"
                                elif len(lookup_keys) == 2:
                                    lookup_key = lookup_keys[1]
                                elif len(lookup_keys) == 0:
                                    raise ValueError(f"A json key cannot be __not. It was : {json_chain_keys[-1]}")
                                else:
                                    raise ValueError(
                                        "Found several lookup parameters, but only one lookup param + an optionnal "
                                        f"'__not' are allowed. The problematic syntax was : {json_chain_keys[-1]}"
                                    )
                                _logger.debug(f"Filtering on lookup json field : {lookup_key}")

                            # we use lookup_key just to know how to compose the query string, in the subsequent if check
                            # but we do not inject it in there, as it should already be present if we find it,
                            # and if not, "exact" will be used by the django backend by default

                            # if lookup key is contains or icontains, and we require several values (list or tuple)
                            # we must make a separate query filter string for each required value in json_value
                            if "contains" in lookup_key and isinstance(json_value, (list, tuple)):
                                for item in json_value:
                                    json_query = f"{'__'.join(json_chain_keys)},{item}"
                                    values.append(json_query)
                            else:
                                json_query = f"{'__'.join(json_chain_keys)},{json_value}"
                                values.append(json_query)

                        # multiple json queries are grouped with a ; They will be splitted in with the same char in the
                        # backend, so the dict keys and values must not contain ; or it will break
                        value = ";".join(values)

                    query_params.append((key, ",".join(map(str, ensure_list(value)))))

                # the ",".join(map(str system allows to convert all lists in query params to comma separated string
                # list if value contains multiple elements

                url = update_url_listparams(url, query_params)
            return self.get("/" + url, **cache_args)
        if not isinstance(id, str) and id is not None:
            id = str(id)  # e.g. may be uuid.UUID
        if action == "read":
            assert endpoint_scheme[action]["action"] == "get"
            return self.get("/" + endpoint + "/" + id.split("/")[-1], **cache_args)
        elif action == "create":
            assert endpoint_scheme[action]["action"] == "post"
            return self.post("/" + endpoint, data=data, files=files)
        elif action == "delete":
            assert endpoint_scheme[action]["action"] == "delete"
            return self.delete("/" + endpoint + "/" + id.split("/")[-1])
        elif action == "partial_update":
            assert endpoint_scheme[action]["action"] == "patch"
            return self.patch("/" + endpoint + "/" + id.split("/")[-1], data=data, files=files)
        elif action == "update":
            assert endpoint_scheme[action]["action"] == "put"
            return self.put("/" + endpoint + "/" + id.split("/")[-1], data=data, files=files)

    # JSON field interface convenience methods
    def _check_inputs(self, endpoint: str) -> None:
        # make sure the queried endpoint exists, if not throw an informative error
        if endpoint not in self.rest_schemes.keys():
            av = [k for k in self.rest_schemes.keys() if not k.startswith("_") and k]
            raise ValueError(
                'REST endpoint "'
                + endpoint
                + '" does not exist. Available '
                + "endpoints are \n       "
                + "\n       ".join(av)
            )
        return

    def json_field_write(
        self,
        endpoint: str = None,
        uuid: str = None,
        field_name: str = None,
        data: dict = None,
    ) -> dict:
        """
        Write data to JSON field.  WILL NOT CHECK IF DATA EXISTS
        NOTE: Destructive write!

        Parameters
        ----------
        endpoint : str, None
            Valid alyx endpoint, defaults to None
        uuid : str, uuid.UUID, None
            UUID or lookup name for endpoint
        field_name : str, None
            Valid json field name, defaults to None
        data : dict, None
            Data to write to json field, defaults to None

        Returns
        -------
        dict
            Written data dict
        """
        self._check_inputs(endpoint)
        # Prepare data to patch
        patch_dict = {field_name: data}
        # Upload new extended_qc to session
        ret = self.rest(endpoint, "partial_update", id=uuid, data=patch_dict)
        return ret[field_name]

    def json_field_update(
        self,
        endpoint: str = None,
        uuid: str = None,
        field_name: str = "json",
        data: dict = None,
    ) -> dict:
        """
        Non-destructive update of JSON field of endpoint for object
        Will update the field_name of the object with pk = uuid of given endpoint
        If data has keys with the same name of existing keys it will squash the old
        values (uses the dict.update() method).

        Parameters
        ----------
        endpoint : str
            Alyx REST endpoint to hit
        uuid : str, uuid.UUID
            UUID or lookup name of object
        field_name : str
            Name of the json field
        data : dict
            A dictionary with fields to be updated

        Returns
        -------
        dict
            New patched json field contents as dict

        Examples
        --------
        >>> client = AlyxClient()
        >>> client.json_field_update("sessions", "eid_str", "extended_qc", {"key": "value"})
        """
        self._check_inputs(endpoint)
        # Load current json field contents
        current = self.rest(endpoint, "read", id=uuid)[field_name]
        if current is None:
            current = {}

        if not isinstance(current, dict):
            _logger.warning(f"Current json field {field_name} does not contains a dict, aborting update")
            return current

        # Patch current dict with new data
        current.update(data)
        # Prepare data to patch
        patch_dict = {field_name: current}
        # Upload new extended_qc to session
        ret = self.rest(endpoint, "partial_update", id=uuid, data=patch_dict)
        return ret[field_name]

    def json_field_remove_key(
        self,
        endpoint: str = None,
        uuid: str = None,
        field_name: str = "json",
        key: str = None,
    ) -> Optional[dict]:
        """
        Remove inputted key from JSON field dict and re-upload it to Alyx.
        Needs endpoint, uuid and json field name

        Parameters
        ----------
        endpoint : str
            Endpoint to hit, defaults to None
        uuid : str
            UUID or lookup name for endpoint
        field_name : str
            JSON field name of object, defaults to None
        key : str
            Key name of dictionary inside object, defaults to None

        Returns
        -------
        dict
            New content of json field
        """
        self._check_inputs(endpoint)
        current = self.rest(endpoint, "read", id=uuid)[field_name]
        # If no contents, cannot remove key, return
        if current is None:
            return current
        # if contents are not dict, cannot remove key, return contents
        if isinstance(current, str):
            _logger.warning(f"Cannot remove key {key} content of json field is of type str")
            return None
        # If key not present in contents of json field cannot remove key, return contents
        if current.get(key, None) is None:
            _logger.warning(f"{key}: Key not found in endpoint {endpoint} field {field_name}")
            return current
        _logger.info(f"Removing key from dict: '{key}'")
        current.pop(key)
        # Re-write contents without removed key
        written = self.json_field_write(endpoint=endpoint, uuid=uuid, field_name=field_name, data=current)
        return written

    def json_field_delete(self, endpoint: str = None, uuid: str = None, field_name: str = None) -> None:
        self._check_inputs(endpoint)
        _ = self.rest(endpoint, "partial_update", id=uuid, data={field_name: None})
        return _[field_name]

    def clear_rest_cache(self):
        """Clear all REST response cache files for the base url"""
        for file in self.cache_dir.joinpath(".rest").glob("*"):
            file.unlink()



"""Unit tests for the one.converters module"""
import unittest
from unittest import mock
from pathlib import Path, PurePosixPath, PureWindowsPath
from uuid import UUID
import datetime

import pandas as pd

from one.api import ONE
from one import converters
from one.alf.files import add_uuid_string
from . import util, OFFLINE_ONLY, TEST_DB_2


class TestConverters(unittest.TestCase):
    tempdir = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.tempdir = util.set_up_env()
        # Create ONE object with temp cache dir
        cls.one = ONE(mode="local", cache_dir=cls.tempdir.name)

    def test_to_eid(self):
        expected = "d3372b15-f696-4279-9be5-98f15783b5bb"
        # Path str
        eid = self.one.to_eid("ZFM-01935/2021-02-05/001")
        self.assertEqual(eid, expected)
        # eid
        eid = self.one.to_eid(eid)
        self.assertEqual(eid, expected)
        # Path
        session_path = Path(self.one.cache_dir).joinpath(
            "mainenlab", "Subjects", "ZFM-01935", "2021-02-05", "001", "alf"
        )
        eid = self.one.to_eid(session_path)
        self.assertEqual(eid, expected)
        # exp ref str
        eid = self.one.to_eid("2021-02-05_001_ZFM-01935")
        self.assertEqual(eid, expected)
        # UUID
        eid = self.one.to_eid(UUID(eid))
        self.assertEqual(eid, expected)
        # session URL
        base_url = "https://alyx.internationalbrainlab.org/"
        eid = self.one.to_eid(base_url + "sessions/" + eid)
        self.assertEqual(eid, expected)
        # None
        self.assertIsNone(self.one.to_eid(None))

        # Test value errors
        with self.assertRaises(ValueError):
            self.one.to_eid("fakeid")
        with self.assertRaises(ValueError):
            self.one.to_eid(util)

    def test_path2eid(self):
        """Test for ConversionMixin.path2eid (offline mode)"""
        verifiable = self.one.path2eid("CSK-im-007/2021-03-21/002")
        self.assertIsNone(verifiable)

        verifiable = self.one.path2eid("ZFM-01935/2021-02-05/001")
        expected = "d3372b15-f696-4279-9be5-98f15783b5bb"
        self.assertEqual(verifiable, expected)

        session_path = Path.home().joinpath("mainenlab", "Subjects", "ZFM-01935", "2021-02-05", "001", "alf")
        verifiable = self.one.path2eid(session_path)
        self.assertEqual(verifiable, expected)

        # Test short circuit
        with mock.patch.object(self.one._cache, "sessions", new=pd.DataFrame([])):
            self.assertIsNone(self.one.path2eid(session_path))

    def test_eid2path(self):
        """Test for ConversionMixin.eid2path"""
        eid = "d3372b15-f696-4279-9be5-98f15783b5bb"
        verifiable = self.one.eid2path(eid)
        expected = Path(self.tempdir.name).joinpath("mainenlab", "Subjects", "ZFM-01935", "2021-02-05", "001")
        self.assertEqual(expected, verifiable)

        with self.assertRaises(ValueError):
            self.one.eid2path("fakeid")
        self.assertIsNone(self.one.eid2path(eid.replace("d", "b")))

        # Test list
        verifiable = self.one.eid2path([eid, eid])
        self.assertIsInstance(verifiable, list)
        self.assertTrue(len(verifiable) == 2)

        # Test short circuit
        with mock.patch.object(self.one._cache, "sessions", new=pd.DataFrame([])):
            self.assertIsNone(self.one.eid2path(eid))

        # Test with int ids
        old_cache = self.one._cache.copy()
        util.caches_str2int(self.one._cache)
        try:
            verifiable = self.one.eid2path(eid)
            expected = Path(self.tempdir.name).joinpath("mainenlab", "Subjects", "ZFM-01935", "2021-02-05", "001")
            self.assertEqual(verifiable, expected)
        finally:
            self.one._cache = old_cache

    def test_eid2ref(self):
        eid = "d3372b15-f696-4279-9be5-98f15783b5bb"
        verifiable = self.one.eid2ref(eid, parse=False, as_dict=True)
        expected = {"subject": "ZFM-01935", "date": "2021-02-05", "sequence": "001"}
        self.assertCountEqual(expected, verifiable)
        verifiable = self.one.eid2ref(eid, parse=True, as_dict=True)
        expected = {"subject": "ZFM-01935", "date": datetime.date(2021, 2, 5), "sequence": 1}
        self.assertCountEqual(expected, verifiable)
        verifiable = self.one.eid2ref(eid, parse=False, as_dict=False)
        self.assertCountEqual("2021-02-05_001_ZFM-01935", verifiable)
        verifiable = self.one.eid2ref(eid, parse=True, as_dict=False)
        self.assertCountEqual("2021-02-05_1_ZFM-01935", verifiable)

    def test_path2record(self):
        """Tests for ConversionMixin.path2record method"""
        file = Path(self.tempdir.name).joinpath(
            "cortexlab", "Subjects", "KS005", "2019-04-02", "001", "alf", "_ibl_wheel.position.npy"
        )
        rec = self.one.path2record(file)
        self.assertIsInstance(rec, pd.Series)
        self.assertTrue(file.as_posix().endswith(rec["rel_path"]))

        # Test URL
        parts = add_uuid_string(file, "94285bfd-7500-4583-83b1-906c420cc667").parts[-7:]
        url = TEST_DB_2["base_url"] + "/".join(("", *parts))
        rec = self.one.path2record(url)
        self.assertIsInstance(rec, pd.Series)
        self.assertTrue(file.as_posix().endswith(rec["rel_path"]))

        file = file.parent / "_fake_obj.attr.npy"
        self.assertIsNone(self.one.path2record(file))

        # Test short circuit
        empty = self.one._cache.datasets.iloc[0:0].copy()
        with mock.patch.object(self.one._cache, "datasets", new=empty):
            self.assertIsNone(self.one.path2record(file))

        # Test empty session
        with mock.patch.object(self.one._cache, "datasets", new=self.one._cache.datasets.iloc[:2]):
            self.assertIsNone(self.one.path2record(file))

        # Test session path input
        session_path = Path(self.tempdir.name).joinpath("cortexlab", "Subjects", "KS005", "2019-04-02", "001")
        rec = self.one.path2record(session_path)
        self.assertIsInstance(rec, pd.Series)
        self.assertEqual(rec.name, "bc93a3b2-070d-47a8-a2b8-91b3b6e9f25c")

        with mock.patch.object(self.one._cache, "sessions", new=empty):
            self.assertIsNone(self.one.path2record(session_path))

    def test_is_exp_ref(self):
        ref = {"date": datetime.datetime(2018, 7, 13).date(), "sequence": 1, "subject": "flowers"}
        self.assertTrue(self.one.is_exp_ref(ref))
        self.assertTrue(self.one.is_exp_ref("2018-07-13_001_flowers"))
        self.assertTrue(self.one.is_exp_ref("2018-07-13_1_flowers"))
        self.assertFalse(self.one.is_exp_ref("2018-invalid_ref-s"))
        # Test recurse
        refs = ("2018-07-13_001_flowers", "2018-07-13_1_flowers")
        self.assertTrue(all(x is True for x in self.one.is_exp_ref(refs)))

    def test_ref2dict(self):
        # Test ref string (none padded)
        d = self.one.ref2dict("2018-07-13_1_flowers")
        expected = {"date": datetime.date(2018, 7, 13), "sequence": 1, "subject": "flowers"}
        self.assertEqual(d, expected)
        # Test short circuit
        self.assertEqual(self.one.ref2dict(d), expected)
        # Test padded number and parse
        d = self.one.ref2dict("2018-07-13_001_flowers", parse=False)
        expected = {"date": "2018-07-13", "sequence": "001", "subject": "flowers"}
        self.assertEqual(d, expected)
        # Test list input
        d = self.one.ref2dict(["2018-07-13_1_flowers", "2020-01-23_002_ibl_witten_01"])
        expected = [
            {"date": datetime.date(2018, 7, 13), "sequence": 1, "subject": "flowers"},
            {"date": datetime.date(2020, 1, 23), "sequence": 2, "subject": "ibl_witten_01"},
        ]
        self.assertEqual(d, expected)

    def test_dict2ref(self):
        d1 = {"date": "2018-07-13", "sequence": "001", "subject": "flowers"}
        self.assertEqual("2018-07-13_001_flowers", self.one.dict2ref(d1))
        d2 = {"date": datetime.date(2018, 7, 13), "sequence": 1, "subject": "flowers"}
        self.assertEqual("2018-07-13_1_flowers", self.one.dict2ref(d2))
        self.assertIsNone(self.one.dict2ref({}))
        expected = ["2018-07-13_001_flowers", "2018-07-13_1_flowers"]
        self.assertCountEqual(expected, self.one.dict2ref([d1, d2]))

    def test_path2ref(self):
        path_str = Path("E:/FlatIron/Subjects/zadorlab/flowers/2018-07-13/001")
        ref = self.one.path2ref(path_str)
        expected = {"subject": "flowers", "date": datetime.date(2018, 7, 13), "sequence": 1}
        self.assertEqual(expected, ref)
        ref = self.one.path2ref(path_str, as_dict=False)
        expected = "2018-07-13_001_flowers"
        self.assertEqual(expected, ref)
        expected = {"subject": "flowers", "date": "2018-07-13", "sequence": "001"}
        ref = self.one.path2ref(path_str, parse=False)
        self.assertEqual(expected, ref)
        path_str2 = "E:/FlatIron/Subjects/churchlandlab/CSHL046/2020-06-20/002"
        refs = self.one.path2ref([path_str, path_str2])
        expected = [
            {"subject": "flowers", "date": datetime.date(2018, 7, 13), "sequence": 1},
            {"subject": "CSHL046", "date": datetime.date(2020, 6, 20), "sequence": 2},
        ]
        self.assertCountEqual(expected, refs)

    def test_ref2path(self):
        ref = {"subject": "flowers", "date": datetime.datetime(2018, 7, 13).date(), "sequence": 1}
        path = self.one.ref2path(ref)
        self.assertIsInstance(path, Path)
        self.assertTrue(path.as_posix().endswith("zadorlab/Subjects/flowers/2018-07-13/001"))
        paths = self.one.ref2path(["2018-07-13_1_flowers", "2019-04-11_1_KS005"])
        expected = ["zadorlab/Subjects/flowers/2018-07-13/001", "cortexlab/Subjects/KS005/2019-04-11/001"]
        self.assertTrue(all(x.as_posix().endswith(y) for x, y in zip(paths, expected)))


@unittest.skipIf(OFFLINE_ONLY, "online only tests")
class TestOnlineConverters(unittest.TestCase):
    """Currently these methods hit the /docs endpoint"""

    @classmethod
    def setUpClass(cls) -> None:
        # Create ONE object with temp cache dir
        cls.one = ONE(**TEST_DB_2)
        cls.eid = "4ecb5d24-f5cc-402c-be28-9d0f7cb14b3a"
        cls.session_record = cls.one.get_details(cls.eid)

    def test_to_eid(self):
        """Test for ConversionMixin.to_eid"""
        eid = self.one.to_eid(self.session_record)
        self.assertEqual(eid, self.eid)

    def test_record2url(self):
        """Test for ConversionMixin.record2url"""
        rec = self.one.get_details(self.eid, full=True, query_type="local")
        idx = rec.rel_path == "alf/probe00/_phy_spikes_subset.channels.npy"
        # As pd.Series
        url = self.one.record2url(rec[idx].squeeze())
        expected = (
            "https://ibl.flatironinstitute.org/public/hoferlab/Subjects/"
            "SWC_043/2020-09-21/001/alf/probe00/"
            "_phy_spikes_subset.channels.00c234a3-a4ff-4f97-a522-939d15528a45.npy"
        )
        self.assertEqual(expected, url)
        # As pd.DataFrame
        url = self.one.record2url(rec[idx])
        self.assertEqual([expected], url)
        # Session record
        rec = self.one._cache["sessions"].loc[self.eid]
        url = self.one.record2url(rec)
        expected = "https://ibl.flatironinstitute.org/public/hoferlab/Subjects/SWC_043/2020-09-21/001"
        self.assertEqual(expected, url)

    def test_record2path(self):
        """Test for ConversionMixin.record2path"""
        rec = self.one.get_details(self.eid, full=True, query_type="local")
        # As pd.Series
        alf_path = "hoferlab/Subjects/SWC_043/2020-09-21/001/alf/probe00/_phy_spikes_subset.channels.npy"
        expected = Path(self.one.alyx.cache_dir).joinpath(*alf_path.split("/"))
        path = self.one.record2path(rec.loc[(self.eid, "00c234a3-a4ff-4f97-a522-939d15528a45")])
        self.assertIsInstance(path, Path)
        self.assertEqual(expected, path)
        # As pd.DataFrame
        idx = rec.rel_path == "alf/probe00/_phy_spikes_subset.channels.npy"
        path = self.one.record2path(rec[idx])
        self.assertEqual(expected, path)

    def test_eid2path(self):
        """Test for OneAlyx.eid2path"""
        verifiable = self.one.eid2path(self.eid, query_type="remote")
        expected = Path(self.one.cache_dir).joinpath(
            "hoferlab",
            "Subjects",
            "SWC_043",
            "2020-09-21",
            "001",
        )
        self.assertEqual(expected, verifiable)

        with self.assertRaises(ValueError):
            self.one.eid2path("fakeid", query_type="remote")
        self.assertIsNone(self.one.eid2path(self.eid.replace("d", "b")))

        # Test list
        verifiable = self.one.eid2path([self.eid, self.eid], query_type="remote")
        self.assertIsInstance(verifiable, list)
        self.assertTrue(len(verifiable) == 2)

    def test_path2eid(self):
        """Test for OneAlyx.path2eid"""
        test_path = Path(self.one.cache_dir).joinpath(
            "hoferlab",
            "Subjects",
            "SWC_043",
            "2020-09-21",
            "001",
        )
        verifiable = self.one.path2eid(test_path, query_type="remote")
        self.assertEqual(self.eid, verifiable)
        # Check works with list
        verifiable = self.one.path2eid([test_path, test_path], query_type="remote")
        self.assertEqual([self.eid, self.eid], verifiable)


class TestAlyx2Path(unittest.TestCase):
    dset = {
        "url": "https://alyx.internationalbrainlab.org/datasets/00059298-1b33-429c-a802-fa51bb662d72",
        "name": "channels.localCoordinates.npy",
        "created_by": "nate",
        "created_datetime": "2020-02-07T22:08:08.053982",
        "dataset_type": "channels.localCoordinates",
        "data_format": "npy",
        "collection": "alf/probe00",
        "session": "https://alyx.internationalbrainlab.org/sessions/7cffad38-0f22-4546-92b5-fd6d2e8b2be9",
        "file_size": 6064,
        "hash": "bc74f49f33ec0f7545ebc03f0490bdf6",
        "version": "1.5.36",
        "experiment_number": 1,
        "file_records": [
            {
                "id": "c9ae1b6e-03a6-41c9-9e1b-4a7f9b5cfdbf",
                "data_repository": "ibl_floferlab_SR",
                "data_repository_path": "/mnt/s0/Data/Subjects/",
                "relative_path": "SWC_014/2019-12-11/001/alf/probe00/channels.localCoordinates.npy",
                "data_url": None,
                "exists": True,
            },
            {
                "id": "f434a638-bc61-4695-884e-70fd1e521d60",
                "data_repository": "flatiron_hoferlab",
                "data_repository_path": "/hoferlab/Subjects/",
                "relative_path": "SWC_014/2019-12-11/001/alf/probe00/channels.localCoordinates.npy",
                "data_url": (
                    "https://ibl.flatironinstitute.org/hoferlab/Subjects/SWC_014/2019-12-11/001/"
                    "alf/probe00/channels.localCoordinates.00059298-1b33-429c-a802-fa51bb662d72.npy"
                ),
                "exists": True,
            },
        ],
        "auto_datetime": "2021-02-10T20:24:31.484939",
    }

    def test_dsets_2_path(self):
        one_path = "/one_root/hoferlab/Subjects/SWC_014/2019-12-11/001/alf/probe00/channels.localCoordinates.npy"

        # Test one_path_from_dataset
        root = PurePosixPath("/one_root")
        testable = converters.one_path_from_dataset(self.dset, one_cache=root)
        self.assertEqual(str(testable), one_path)
        # Check list input
        testable = converters.one_path_from_dataset([self.dset], one_cache=root)
        self.assertIsInstance(testable, list)
        # Check handles string inputs
        testable = converters.one_path_from_dataset(self.dset, one_cache="/one_root")
        self.assertTrue(hasattr(testable, "is_absolute"), "Failed to return Path object")
        self.assertEqual(str(testable).replace("\\", "/"), one_path)

        # Test one_path_from_dataset using Windows path
        one_path = PureWindowsPath(r"C:/Users/User/")
        testable = converters.one_path_from_dataset(self.dset, one_cache=one_path)
        self.assertIsInstance(testable, PureWindowsPath)
        self.assertTrue(str(testable).startswith(str(one_path)))
        self.assertTrue("hoferlab/Subjects" in testable.as_posix())
        # Check repository arg
        testable = converters.path_from_dataset(self.dset, root_path=root, repository="ibl_floferlab_SR")
        self.assertTrue("mnt/s0/Data/Subjects" in testable.as_posix())

        # Tests path_from_filerecord: when given a string, a system path object should be returned
        fr = self.dset["file_records"][0]
        testable = converters.path_from_filerecord(fr, root_path="C:\\")
        self.assertIsInstance(testable, Path)
        # Check list
        testable = converters.path_from_filerecord([fr], root_path="C:\\")
        self.assertIsInstance(testable, list)
        # Check uuid
        uuid = "00059298-1b33-429c-a802-fa51bb662d72"
        testable = converters.path_from_filerecord(fr, root_path="C:\\", uuid=uuid)
        self.assertTrue(uuid in testable.as_posix())

    def test_session_record2path(self):
        """Test for one.converters.session_record2path"""
        rec = {"subject": "ALK01", "date": "2020-01-01", "number": 1}
        path = converters.session_record2path(rec)
        self.assertEqual(path, PurePosixPath("ALK01/2020-01-01/001"))

        rec = {
            "date": datetime.datetime.fromisoformat("2020-01-01").date(),
            "number": "001",
            "lab": "foo",
            "subject": "ALK01",
        }
        path = converters.session_record2path(rec, str(Path.home()))
        self.assertEqual(path, Path.home() / "foo/Subjects/ALK01/2020-01-01/001")


class TestWrappers(unittest.TestCase):
    def test_recurse(self):
        # Check accepts different numbers of input args
        wrapped = converters.recurse(lambda x, y, z: y * 2)
        self.assertEqual(wrapped(1, 2, 3), 4)
        wrapped = converters.recurse(lambda x, y: y * 2)
        self.assertEqual(wrapped(1, 2), 4)
        wrapped = converters.recurse(lambda: 8)
        self.assertEqual(wrapped(), 8)
        # Check recurse of lists/tuples
        wrapped = converters.recurse(lambda x, y: y * 2)
        self.assertEqual(wrapped(None, [1, 2, 3, 4]), [2, 4, 6, 8])
        self.assertEqual(wrapped(None, (1, 2, 3, 4)), [2, 4, 6, 8])

    def test_parse_values(self):
        wrapped = converters.parse_values(lambda x: x)
        # Check ignores when str
        ref = "subject_1_2021-01-01"
        self.assertEqual(ref, wrapped(ref, parse=True))
        # Check parse values false (should be identity)
        ref = {"subject": "flowers", "sequence": "001", "date": "2021-01-01"}
        self.assertEqual(ref, wrapped(ref.copy(), parse=False))
        # Check parse values true (should change sequence and date)
        expected = {"subject": "flowers", "sequence": 1, "date": datetime.date(2021, 1, 1)}
        self.assertEqual(expected, wrapped(ref.copy(), parse=True))
        # Check handles datetime ISO
        ref["date"] += "T14:53:53.586024"
        self.assertEqual(expected, wrapped(ref.copy(), parse=True))
        # Check handles list
        self.assertIsInstance(wrapped([ref.copy(), expected], parse=True), list)


if __name__ == "__main__":
    unittest.main(exit=False)



"""ALyx File related errors

A set of Alyx and ALF related error classes which provide a more verbose description of the raised
issues.
"""


class ALFError(Exception):
    """A base class for ALF-related errors

    Attributes
    ----------
    explanation : str
        An optional, verbose but general explanation of the error class.  All errors will display
        the same explanation.
    """
    explanation = ''

    def __init__(self, *args, terse=False):
        """A base ALF exception

        Parameters
        ----------
        args : str, any
            A specific error message to display or items to include in message
        terse : bool
            If True, the explanation string is not included in exception message

        Examples
        --------
        >>> raise ALFError("ALF directory doesn't exist")
        one.alf.exceptions.ALFError: ALF directory doesn't exist

        >>> raise ALFError('invalid/path/one', 'invalid/path/two')
        one.alf.exceptions.ALFError: "invalid/path/one", "invalid/path/two"
        """
        if args:
            if len(args) == 1 and isinstance(args[0], str):
                self.message = args[0]
            else:
                self.message = '"' + '", "'.join(map(str, args)) + '"'
        else:
            self.message = ''
        self.terse = terse

    def __str__(self):
        if not self.message and not self.explanation:
            return ''
        return self.message if self.terse else f'{self.message} \n {self.explanation} '


class AlyxSubjectNotFound(ALFError):
    """'Subject not found' error"""
    explanation = 'The subject was not found in Alyx database'


class ALFObjectNotFound(ALFError):
    """'Object not found' error"""
    explanation = ('The ALF object was not found.  This may occur if the object or namespace or '
                   'incorrectly formatted e.g. the object "_ibl_trials.intervals.npy" would be '
                   'found with the filters `object="trials", namespace="ibl"`')


class ALFMultipleObjectsFound(ALFError):
    """'Multiple objects found' error"""
    explanation = ('Dataset files belonging to more than one object found.  '
                   'ALF names have the pattern '
                   '(_namespace_)object.attribute(_timescale).extension, e.g. for the file '
                   '"_ibl_trials.intervals.npy" the object is "trials"')


class ALFMultipleCollectionsFound(ALFError):
    """'Multiple collections found' error"""
    explanation = ('The matching object/file(s) belong to more than one collection.  '
                   'ALF names have the pattern '
                   'collection/(_namespace_)object.attribute(_timescale).extension, e.g. for the '
                   'file "alf/probe01/spikes.times.npy" the collection is "alf/probe01"')


class ALFMultipleRevisionsFound(ALFError):
    """'Multiple objects found' error"""
    explanation = ('The matching object/file(s) belong to more than one revision.  '
                   'Multiple datasets in different revision folders were found with no default '
                   'specified.')



"""
Module for identifying and parsing ALF file names.

An ALF file has the following components (those in brackets are optional):
    (_namespace_)object.attribute(_timescale)(.extra.parts).ext

Note the following:
    Object attributes may not contain an underscore unless followed by 'times' or 'intervals'.
    A namespace must not contain extra underscores (i.e. `name_space` and `__namespace__` are not
    valid)
    ALF files must always have an extension

For more information, see the following documentation:
    https://int-brain-lab.github.io/ONE/alf_intro.html

Created on Tue Sep 11 18:06:21 2018

@author: Miles
"""
from collections import OrderedDict
from datetime import datetime
from typing import Union, Optional
from pathlib import Path
import logging

from . import spec
from .spec import (
    SESSION_SPEC,
    COLLECTION_SPEC,
    FILE_SPEC,
    REL_PATH_SPEC,
    FULL_ABSOLUTE_SPEC,
    SESSION_ABSOLUTE_SPEC,
    FOLDER_SPEC,
    FULL_FOLDER_SPEC,
)

_logger = logging.getLogger(__name__)


def rel_path_parts(rel_path, as_dict=False, assert_valid=True) -> [dict, tuple]:
    """Parse a relative path into the relevant parts.

    A relative path follows the pattern
        (collection/)(#revision#/)_namespace_object.attribute_timescale.extra.extension

    Parameters
    ----------
    rel_path : str, pathlib.Path
        A relative path string
    as_dict : bool
        If true, an OrderedDict of parts are returned with the keys ('lab', 'subject', 'date',
        'number'), otherwise a tuple of values are returned
    assert_valid : bool
        If true a ValueError is raised when the session cannot be parsed, otherwise an empty
        dict of tuple of Nones is returned

    Returns
    -------
    OrderedDict, tuple
        A dict if as_dict is true, or a tuple of parsed values
    """
    return _path_parts(rel_path, REL_PATH_SPEC, True, as_dict, assert_valid)


def absolute_full_path_parts(session_path, as_dict=False, assert_valid=True):
    """_summary_

    Args:
        session_path (_type_): _description_
        as_dict (bool, optional): _description_. Defaults to False.
        assert_valid (bool, optional): _description_. Defaults to True.

    Return keys :
        - 'root'
        - 'lab'
        - 'subject'
        - 'date'
        - 'number'
        - 'collection'
        - 'revision'
        - 'namespace'
        - 'object'
        - 'attribute'
        - 'timescale'
        - 'extra'
        - 'extension'
    """
    return _path_parts(session_path, FULL_ABSOLUTE_SPEC, False, as_dict, assert_valid)


def session_path_parts(session_path, as_dict=False, assert_valid=True, absolute=False):
    """Parse a session path into the relevant parts.

    Return keys:
        - lab
        - subject
        - date
        - number
        Also includes root if absolute is True

    Parameters
    ----------
    session_path : str, pathlib.Path
        A session path string
    as_dict : bool
        If true, an OrderedDict of parts are returned with the keys ('lab', 'subject', 'date',
        'number'), otherwise a tuple of values are returned
    assert_valid : bool
        If true a ValueError is raised when the session cannot be parsed, otherwise an empty
        dict of tuple of Nones is returned

    Returns
    -------
    OrderedDict, tuple
        A dict if as_dict is true, or a tuple of parsed values

    Raises
    ------
    ValueError
        Invalid ALF session path (assert_valid is True)
    """
    if absolute:
        spec_pattern = SESSION_ABSOLUTE_SPEC
    else:
        spec_pattern = SESSION_SPEC

    return _path_parts(session_path, spec_pattern, False, as_dict, assert_valid)


def _path_parts(path, spec_str, match=True, as_dict=False, assert_valid=True):
    """Given a ALF and a spec string, parse into parts

    Parameters
    ----------
    path : str, pathlib.Path
        An ALF path or dataset
    match : bool
        If True, string must match exactly, otherwise search for expression within path
    as_dict : bool
        When true a dict of matches is returned
    assert_valid : bool
        When true an exception is raised when the filename cannot be parsed

    Returns
    -------
    OrderedDict, tuple
        A dict if as_dict is true, or a tuple of parsed values

    Raises
    ------
    ValueError
        Invalid ALF path (assert_valid is True)
    """
    if hasattr(path, "as_posix"):
        path = path.as_posix()
    pattern = spec.regex(spec_str)
    empty = OrderedDict.fromkeys(pattern.groupindex.keys())
    parsed = (pattern.match if match else pattern.search)(path)
    if parsed:  # py3.8
        parsed_dict = parsed.groupdict()
        return OrderedDict(parsed_dict) if as_dict else tuple(parsed_dict.values())
    elif assert_valid:
        raise ValueError(f'Invalid ALF: "{path}"')
    else:
        return empty if as_dict else tuple(empty.values())


def filename_parts(filename, as_dict=False, assert_valid=True) -> Union[dict, tuple]:
    """
    Return the parsed elements of a given ALF filename.

    Parameters
    ----------
    filename : str
        The name of the file
    as_dict : bool
        When true a dict of matches is returned
    assert_valid : bool
        When true an exception is raised when the filename cannot be parsed

    Returns
    -------
    namespace : str
        The _namespace_ or None if not present
    object : str
        ALF object
    attribute : str
        The ALF attribute
    timescale : str
        The ALF _timescale or None if not present
    extra : str
        Any extra parts to the filename, or None if not present
    extension : str
        The file extension

    Examples
    --------
    >>> filename_parts('_namespace_obj.times_timescale.extra.foo.ext')
    ('namespace', 'obj', 'times', 'timescale', 'extra.foo', 'ext')
    >>> filename_parts('spikes.clusters.npy', as_dict=True)
    {'namespace': None,
     'object': 'spikes',
     'attribute': 'clusters',
     'timescale': None,
     'extra': None,
     'extension': 'npy'}
    >>> filename_parts('spikes.times_ephysClock.npy')
    (None, 'spikes', 'times', 'ephysClock', None, 'npy')
    >>> filename_parts('_iblmic_audioSpectrogram.frequencies.npy')
    ('iblmic', 'audioSpectrogram', 'frequencies', None, None, 'npy')
    >>> filename_parts('_spikeglx_ephysData_g0_t0.imec.wiring.json')
    ('spikeglx', 'ephysData_g0_t0', 'imec', None, 'wiring', 'json')
    >>> filename_parts('_spikeglx_ephysData_g0_t0.imec0.lf.bin')
    ('spikeglx', 'ephysData_g0_t0', 'imec0', None, 'lf', 'bin')
    >>> filename_parts('_ibl_trials.goCue_times_bpod.csv')
    ('ibl', 'trials', 'goCue_times', 'bpod', None, 'csv')

    Raises
    ------
    ValueError
        Invalid ALF dataset (assert_valid is True)
    """
    return _path_parts(filename, FILE_SPEC, True, as_dict, assert_valid)


def full_path_parts(
    path, as_dict=False, assert_valid=True, absolute=False
) -> Union[dict, tuple]:
    """Parse all filename and folder parts.

    Parameters
    ----------
    path : str, pathlib.Path
        The ALF path
    as_dict : bool
        When true a dict of matches is returned
    assert_valid : bool
        When true an exception is raised when the filename cannot be parsed

    Returns
    -------
    OrderedDict, tuple
        A dict if as_dict is true, or a tuple of parsed values

    Examples
    --------
    >>> full_path_parts(
    ...    'lab/Subjects/subject/2020-01-01/001/collection/#revision#/'
    ...    '_namespace_obj.times_timescale.extra.foo.ext')
    ('lab', 'subject', '2020-01-01', '001', 'collection', 'revision',
    'namespace', 'obj', 'times','timescale', 'extra.foo', 'ext')
    >>> full_path_parts('spikes.clusters.npy', as_dict=True)
    {'lab': None,
     'subject': None,
     'date': None,
     'number': None,
     'collection': None,
     'revision': None,
     'namespace': None,
     'object': 'spikes',
     'attribute': 'clusters',
     'timescale': None,
     'extra': None,
     'extension': 'npy'}

    Raises
    ------
    ValueError
        Invalid ALF path (assert_valid is True)
    """
    path = Path(path)
    # NB We try to determine whether we have a folder or filename path.  Filenames contain at
    # least two periods, however it is currently permitted to have any number of periods in a
    # collection, making the ALF path ambiguous.
    # if sum(x == '.' for x in path.name) < 2:  # folder only
    #    folders = folder_parts(path, as_dict, assert_valid, absolute = absolute)
    #    dataset = filename_parts('', as_dict, assert_valid=False)
    if "/" not in path.as_posix() and "\\" not in path.as_posix():  # filename only
        folders = folder_parts("", as_dict, assert_valid=False, absolute=absolute)
        dataset = filename_parts(path.name, as_dict, assert_valid)
    else:  # full filepath
        folders = folder_parts(path.parent, as_dict, assert_valid, absolute=absolute)
        dataset = filename_parts(path.name, as_dict, assert_valid)
    if as_dict:
        return {
            key: value if value is not None else ""
            for key, value in OrderedDict(**folders, **dataset).items()
        }
    else:
        return folders + dataset


def folder_parts(
    folder_path, as_dict=False, assert_valid=True, absolute=False
) -> Union[dict, tuple]:
    """Parse all folder parts, including session, collection and revision.

    Parameters
    ----------
    folder_path : str, pathlib.Path
        The ALF folder path
    as_dict : bool
        When true a dict of matches is returned
    assert_valid : bool
        When true an exception is raised when the filename cannot be parsed

    Returns
    -------
    OrderedDict, tuple
        A dict if as_dict is true, or a tuple of parsed values

    Examples
    --------
    >>> folder_parts('lab/Subjects/subject/2020-01-01/001/collection/#revision#')
    ('lab', 'subject', '2020-01-01', '001', 'collection', 'revision')
    >>> folder_parts(Path('lab/Subjects/subject/2020-01-01/001'), as_dict=True)
    {'lab': 'lab',
     'subject': 'subject',
     'date': '2020-01-01',
     'number': '001',
     'collection': None,
     'revision': None}

    Raises
    ------
    ValueError
        Invalid ALF path (assert_valid is True)
    """
    if hasattr(folder_path, "as_posix"):
        folder_path = folder_path.as_posix()
    if folder_path and folder_path[-1] != "/":  # Slash required for regex pattern
        folder_path = folder_path + "/"
    if absolute:
        spec_str = FULL_FOLDER_SPEC + "$"
    else:
        spec_str = FOLDER_SPEC + "$"
    return _path_parts(folder_path, spec_str, False, as_dict, assert_valid)


def _isdatetime(s: str) -> bool:
    """Returns True if input is valid ISO date string"""
    try:
        datetime.strptime(s, "%Y-%m-%d")
        return True
    except ValueError:
        return False


def get_session_path(path: Union[str, Path]) -> Optional[Path]:
    """
    Returns the session path from any filepath if the date/number pattern is found,
    including the root directory

    Returns
    -------
    pathlib.Path
        The session path part of the input path or None if path invalid.

    Examples
    --------
    >>> get_session_path('/mnt/sd0/Data/lab/Subjects/subject/2020-01-01/001')
    Path('/mnt/sd0/Data/lab/Subjects/subject/2020-01-01/001')

    >>> get_session_path('C:\\Data\\subject\\2020-01-01\\1\\trials.intervals.npy')
    Path('C:/Data/subject/2020-01-01/1')
    """
    if path is None:
        return
    if isinstance(path, str):
        path = Path(path)
    sess = None
    for i, p in enumerate(path.parts):
        if p.isdigit() and _isdatetime(path.parts[i - 1]):
            sess = Path().joinpath(*path.parts[: i + 1])

    return sess


def get_alf_path(path: Union[str, Path]) -> str:
    """Returns the ALF part of a path or filename.
    Attempts to return the first valid part of the path, first searching for a session path,
    then relative path (collection/revision/filename), then just the filename.  If all invalid,
    None is returned.

    Parameters
    ----------
    path : str, pathlib.Path
        A path to parse

    Returns
    -------
    str
        A string containing the full ALF path, session path, relative path or filename

    Examples
    --------
    >>> get_alf_path('etc/etc/lab/Subjects/subj/2021-01-21/001')
    'lab/Subjects/subj/2021-01-21/001/collection/file.attr.ext'

    >>> get_alf_path('etc/etc/subj/2021-01-21/001/collection/file.attr.ext')
    'subj/2021-01-21/001/collection/file.attr.ext'

    >>> get_alf_path('collection/file.attr.ext')
    'collection/file.attr.ext'
    """
    if not isinstance(path, str):
        path = Path(path).as_posix()
    path = path.strip("/")

    # Check if session path
    match_session = spec.regex(SESSION_SPEC).search(path)
    if match_session:
        return path[match_session.start() :]

    # Check if filename / relative path (i.e. collection + filename)
    parts = path.rsplit("/", 1)
    match_filename = spec.regex(FILE_SPEC).match(parts[-1])
    if match_filename:
        return (
            path
            if spec.regex(f"{COLLECTION_SPEC}{FILE_SPEC}").match(path)
            else parts[-1]
        )


def add_uuid_string(file_path, uuid):
    """
    Add a UUID and an extra part to the filename of an ALF path

    Parameters
    ----------
    file_path : str, pathlib.Path, pathlib.PurePath
        An ALF path to add the UUID to
    uuid : str, uuid.UUID
        The UUID to add

    Returns
    -------
    pathlib.Path
        A new Path object with a UUID in the filename
    """
    if isinstance(uuid, str) and not spec.is_uuid_string(uuid):
        raise ValueError("Should provide a valid UUID v4")
    uuid = str(uuid)
    # NB: Only instantiate as Path if not already a Path, otherwise we risk changing the class
    if isinstance(file_path, str):
        file_path = Path(file_path)
    name_parts = file_path.stem.split(".")
    if uuid == name_parts[-1]:
        _logger.warning(f"UUID already found in file name: {file_path.name}: IGNORE")
        return file_path
    return file_path.parent.joinpath(f"{'.'.join(name_parts)}.{uuid}{file_path.suffix}")



"""Unit tests for Alyx REST queries using the AlyxClient.rest method"""
from pathlib import Path
import unittest
import random
import string
from uuid import UUID

import numpy as np

from one.api import ONE
from one.tests import TEST_DB_1, OFFLINE_ONLY


one = ONE(**TEST_DB_1)

EID = 'cf264653-2deb-44cb-aa84-89b82507028a'
EID_EPHYS = 'b1c968ad-4874-468d-b2e4-5ffa9b9964e9'


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class Tests_REST(unittest.TestCase):
    """Specific Alyx REST endpoint tests"""
    def test_water_restriction(self):
        """
        Test listing water-restriction endpoint.

        Examples of how to list all water restrictions and water-restriction for a given
        subject.
        """
        # get all the water restrictions from start
        all_wr = one.alyx.rest('water-restriction', 'list')
        # 2 different ways to  get water restrictions for one subject
        wr_sub2 = one.alyx.rest('water-restriction', 'list', subject='algernon')  # recommended
        # enforce test logic
        expected = {'end_time', 'reference_weight', 'start_time', 'subject', 'water_type'}
        self.assertTrue(expected >= set(all_wr[0].keys()))
        self.assertTrue(len(all_wr) > len(wr_sub2))

    def test_list_pk_query(self):
        """
        Test REST list with id keyword argument.

        It's a bit stupid but the REST endpoint can't forward a direct query of the uuid via
        the pk keyword. The AlyxClient has already an id parameter, which on the list method
        is used as a pk identifier. This special case is tested here.
        """
        # Sessions returned sorted: take last session as new sessions constantly added and
        # removed by parallel test runs
        ses = one.alyx.rest('sessions', 'list')[-1]
        eid = UUID(ses['url'][-36:])  # Should work with UUID object
        ses_ = one.alyx.rest('sessions', 'list', id=eid)[-1]
        self.assertEqual(ses, ses_)
        # Check works with django query arg
        query = f'start_time__date,{ses["start_time"][:10]}'
        ses_ = one.alyx.rest('sessions', 'list', id=eid, django=query)[-1]
        self.assertEqual(ses, ses_)

    def test_note_with_picture_upload(self):
        """Test adding session note with attached picture"""
        my_note = {'user': 'olivier',
                   'content_type': 'session',
                   'object_id': EID,
                   'text': 'gnagnagna'}

        png = Path(__file__).parent.joinpath('fixtures', 'test_img.png')
        with open(png, 'rb') as img_file:
            files = {'image': img_file}
            ar_note = one.alyx.rest('notes', 'create', data=my_note, files=files)

        self.assertTrue(len(ar_note['image']))
        self.assertTrue(ar_note['content_type'] == 'actions.session')
        one.alyx.rest('notes', 'delete', id=ar_note['id'])

    def test_channels(self):
        """Test creation of insertion, trajectory and channels"""
        # need to build insertion + trajectory + channels to test the serialization of a
        # record array in the channel endpoint
        name = ''.join(random.choices(string.ascii_letters, k=5))
        # Find any existing insertions with this name and delete (unlikely to find any)
        probe_insertions = one.alyx.rest('insertions', 'list',
                                         session=EID_EPHYS, name=name, no_cache=True)
        for pi in probe_insertions:
            one.alyx.rest('insertions', 'delete', pi['id'])
        # Create new insertion with this name and add teardown hook to delete it
        probe_insertion = one.alyx.rest(
            'insertions', 'create', data={'session': EID_EPHYS, 'name': name})
        self.addCleanup(one.alyx.rest, 'insertions', 'delete', id=probe_insertion['id'])
        trajectory = one.alyx.rest('trajectories', 'create', data={
            'probe_insertion': probe_insertion['id'],
            'x': 1500,
            'y': -2000,
            'z': 0,
            'depth': 4500,
            'phi': 0,
            'theta': 0,
            'provenance': 'Histology track',
        })
        channel_records = []
        for _ in np.arange(3):
            channel_records.append({
                'x': np.random.randint(-2000, 2000),
                'y': np.random.randint(-2000, 2000),
                'z': np.random.randint(-2000, 2000),
                'axial': np.random.rand() * 800,
                'lateral': np.random.rand() * 8,
                'brain_region': 889,
                'trajectory_estimate': trajectory['id']
            })
        channels = one.alyx.rest('channels', 'create', data=channel_records)
        self.assertTrue(len(channels) == 3)


if __name__ == "__main__":
    unittest.main(exit=False)
